#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Background thread handler.

Provides a way to create new threads which are not bound to the creator's
request context and do not need to end before the creator request completes.
"""



import logging
import sys
import threading
import traceback

from google.appengine.runtime import request_environment
import six.moves._thread

BACKGROUND_REQUEST_ID = 'HTTP_X_APPENGINE_BACKGROUNDREQUEST'


class _BackgroundRequest(object):
  """A class for coordinating between a background thread and its creator.

  This facilitates swapping target, args and kwargs for thread_id between the
  background thread creator and the background thread.
  """

  def __init__(self):
    self._ready_condition = threading.Condition()
    self._callable_ready = False
    self._thread_id_ready = False

  def ProvideCallable(self, target, args, kwargs):
    """Sets the target and the args to be provided to this background request.

    Args:
      target: A callable for the background thread to execute.
      args: A tuple of positional args to be passed to target.
      kwargs: A dict of keyword args to be passed to target.

    Returns:
      The thread ID of the thread servicing this background request.
    """
    with self._ready_condition:
      self._target = target
      self._args = args
      self._kwargs = kwargs
      self._callable_ready = True
      self._ready_condition.notify()
      while not self._thread_id_ready:
        self._ready_condition.wait()
      return self._thread_id

  def WaitForCallable(self):
    """Sets the thread ID and returns the callable and args for this request.

    This will block until the request details have been set.

    Returns:
      A tuple (target, args, kwargs) where
        target: A callable for the background thread to execute.
        args: A tuple of positional args to be passed to target.
        kwargs: A dict of keyword args to be passed to target.
    """
    with self._ready_condition:
      self._thread_id = six.moves._thread.get_ident()
      self._thread_id_ready = True
      self._ready_condition.notify()
      while not self._callable_ready:
        self._ready_condition.wait()
      return self._target, self._args, self._kwargs


class _BackgroundRequestsContainer(object):
  """A container for storing pending background thread requests."""

  def __init__(self):
    self._requests = {}
    self._lock = threading.Lock()

  def _GetOrAddRequest(self, request_id):
    with self._lock:
      if request_id in self._requests:
        return self._requests[request_id]
      else:
        request = _BackgroundRequest()
        self._requests[request_id] = request
        return request

  def _RemoveRequest(self, request_id):
    with self._lock:
      del self._requests[request_id]

  def EnqueueBackgroundThread(self, request_id, target, args, kwargs):
    """Enqueues a new background thread request for a certain request ID.

    Args:
      request_id: A str containing the request ID for this background thread.
      target: A callable for the background thread to execute.
      args: A tuple of positional args to be passed to target.
      kwargs: A dict of keyword args to be passed to target.

    Returns:
      The thread_id of the background thread.
    """
    request = self._GetOrAddRequest(request_id)
    return request.ProvideCallable(target, args, kwargs)

  def RunBackgroundThread(self, request_id):
    """Runs the callable enqueued for the specified request ID."""
    request = self._GetOrAddRequest(request_id)
    target, args, kwargs = request.WaitForCallable()
    self._RemoveRequest(request_id)
    target(*args, **kwargs)

_pending_background_threads = _BackgroundRequestsContainer()


def EnqueueBackgroundThread(request_id, target, args, kwargs):
  """Enqueues a new background thread request for a certain request ID.

  Args:
    request_id: A str containing the request ID for this background thread.
    target: A callable for the background thread to execute.
    args: A tuple of positional args to be passed to target.
    kwargs: A dict of keyword args to be passed to target.

  Returns:
    The thread_id of the background thread.
  """
  return _pending_background_threads.EnqueueBackgroundThread(
      request_id, target, args, kwargs)


def _HandleNoInit(environ):
  """Handles a background request, skipping request initialization."""

  try:
    request_id = environ[BACKGROUND_REQUEST_ID]
    _pending_background_threads.RunBackgroundThread(request_id)

  except:
    exception = sys.exc_info()


    tb = exception[2].tb_next
    if tb:
      tb = tb.tb_next
    message = ''.join(traceback.format_exception(exception[0], exception[1],
                                                 tb))
    logging.error(message)
    return dict(error=1, response_code=500, message=message)

  return dict(error=0, response_code=200, message='OK')


def Handle(environ):
  """Handles a background request.

  This function is called by the runtime in response to an /_ah/background
  request.

  Args:
    environ: A dict containing the environ for this request (i.e. for
        os.environ).

  Returns:
    A dict containing:
      error: App Engine error code. 0 if completed successfully, 1 otherwise.
      response_code: The HTTP response code. 200 if completed successfully, 500
          otherwise.
      logs: A list of tuples (timestamp_usec, severity, message) of log
          entries.  timestamp_usec is a long, severity is int and message is
          str.  Severity levels are 0..4 for Debug, Info, Warning, Error,
          Critical.
  """
  request_environment.current_request.Init(sys.stderr, environ)
  try:
    response = _HandleNoInit(environ)
  finally:
    request_environment.current_request.Clear()
  return response


def App(environ, start_response):
  """Present Handle() as a WSGI app for Titanoboa."""




  response = _HandleNoInit(environ)
  start_response(str(response['response_code']) + ' ' + response['message'],
                 [('Content-Type', 'text/plain')])
  yield str(response).encode()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Implementation of the datastore_v4 API that forwards to the v3 service."""













from google.appengine.api import api_base_pb2
from google.appengine.api import apiproxy_stub
from google.appengine.api import apiproxy_stub_map
from google.appengine.datastore import datastore_pb
from google.appengine.datastore import datastore_pbs
from google.appengine.datastore import datastore_query
from google.appengine.datastore import datastore_stub_util
from google.appengine.datastore import datastore_v4_pb2
from google.appengine.datastore import datastore_v4_validator
from google.appengine.runtime import apiproxy_errors
from google.appengine.datastore import entity_bytes_pb2 as entity_pb2


SERVICE_NAME = 'datastore_v4'
V3_SERVICE_NAME = 'datastore_v3'


class DatastoreV4Stub(apiproxy_stub.APIProxyStub):
  """Implementation of the datastore_v4 API that forwards to the v3 service."""


  THREADSAFE = False

  def __init__(self, app_id):
    apiproxy_stub.APIProxyStub.__init__(self, SERVICE_NAME)
    self.__app_id = app_id
    self.__entity_converter = datastore_pbs.get_entity_converter()
    self.__service_converter = datastore_stub_util.get_service_converter()
    self.__service_validator = datastore_v4_validator.get_service_validator()

  def _Dynamic_BeginTransaction(self, req, resp):
    try:
      self.__service_validator.validate_begin_transaction_req(req)
      v3_req = self.__service_converter.v4_to_v3_begin_transaction_req(
          self.__app_id, req)
      v3_resp = datastore_pb.Transaction()
      self.__make_v3_call('BeginTransaction', v3_req, v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    try:
      v4_resp = self.__service_converter.v3_to_v4_begin_transaction_resp(
          v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.INTERNAL_ERROR, str(e))
    resp.CopyFrom(v4_resp)

  def _Dynamic_Rollback(self, req, unused_resp):
    try:
      self.__service_validator.validate_rollback_req(req)
      v3_req = self.__service_converter.v4_rollback_req_to_v3_txn(req)
      self.__make_v3_call('Rollback', v3_req, api_base_pb2.VoidProto())
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))

  def _Dynamic_Commit(self, req, resp):
    try:
      self.__service_validator.validate_commit_req(req)
      if req.HasField('transaction'):
        resp.deprecated_mutation_result.Clear()
        resp.deprecated_mutation_result.CopyFrom(
            self.__apply_v4_deprecated_mutation(req.deprecated_mutation,
                                                req.transaction))
        v3_req = self.__service_converter.v4_commit_req_to_v3_txn(req)
        v3_resp = datastore_pb.CommitResponse()
        self.__make_v3_call('Commit', v3_req, v3_resp)
        total_index_updates = (
            resp.deprecated_mutation_result.index_updates
            + v3_resp.cost.index_writes)
        resp.deprecated_mutation_result.index_updates = total_index_updates
      else:
        resp.deprecated_mutation_result.CopyFrom(
            self.__apply_v4_deprecated_mutation(req.deprecated_mutation,
                                                None))
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))

  def _Dynamic_RunQuery(self, req, resp):
    try:
      self.__normalize_v4_run_query_request(req)
      self.__service_validator.validate_run_query_req(req)
      v3_req = self.__service_converter.v4_run_query_req_to_v3_query(req)
      v3_resp = datastore_pb.QueryResult()
      self.__make_v3_call('RunQuery', v3_req, v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    try:
      v4_resp = self.__service_converter.v3_to_v4_run_query_resp(v3_resp)
      if req.query.projection:
        if req.query.projection == ['__key__']:
          result_type = datastore_v4_pb2.EntityResult.KEY_ONLY
        else:
          result_type = datastore_v4_pb2.EntityResult.PROJECTION
        v4_resp.batch.entity_result_type = result_type
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.INTERNAL_ERROR, str(e))
    resp.CopyFrom(v4_resp)

  def _Dynamic_ContinueQuery(self, req, resp):
    try:
      self.__service_validator.validate_continue_query_req(req)
      v3_req = self.__service_converter.v4_to_v3_next_req(req)
      v3_resp = datastore_pb.QueryResult()
      self.__make_v3_call('Next', v3_req, v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    try:
      v4_resp = self.__service_converter.v3_to_v4_continue_query_resp(v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.INTERNAL_ERROR, str(e))
    resp.CopyFrom(v4_resp)

  def _Dynamic_Lookup(self, req, resp):
    try:
      self.__service_validator.validate_lookup_req(req)
      v3_req = self.__service_converter.v4_to_v3_get_req(req)
      v3_resp = datastore_pb.GetResponse()
      self.__make_v3_call('Get', v3_req, v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    try:
      v4_resp = self.__service_converter.v3_to_v4_lookup_resp(v3_resp)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.INTERNAL_ERROR, str(e))
    resp.CopyFrom(v4_resp)

  def _Dynamic_AllocateIds(self, req, resp):



    v3_stub = apiproxy_stub_map.apiproxy.GetStub(V3_SERVICE_NAME)
    try:
      self.__service_validator.validate_allocate_ids_req(req)

      if req.allocate:
        v3_refs = self.__entity_converter.v4_to_v3_references(
            req.allocate)

        v3_full_refs = v3_stub._AllocateIds(v3_refs)
        resp.allocated.extend(
            self.__entity_converter.v3_to_v4_keys(v3_full_refs))
      elif req.reserve:
        v3_refs = self.__entity_converter.v4_to_v3_references(
            req.reserve)

        v3_stub._AllocateIds(v3_refs)
    except datastore_pbs.InvalidConversionError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))
    except datastore_v4_validator.ValidationError as e:
      raise apiproxy_errors.ApplicationError(
          datastore_v4_pb2.Error.BAD_REQUEST, str(e))

  def __insert_v3_entity(self, v3_entity, v3_txn):
    """Inserts a v3 entity.

    Args:
      v3_entity: a datastore_v4_pb2.Entity
      v3_txn: a datastore_pb.Transaction or None

    Returns:
      the number of index writes that occurred

    Raises:
      ApplicationError: if the entity already exists
    """
    if not v3_txn:

      v3_txn = datastore_pb.Transaction()
      v3_begin_txn_req = datastore_pb.BeginTransactionRequest()
      v3_begin_txn_req.app = v3_entity.key.app
      self.__make_v3_call('BeginTransaction', v3_begin_txn_req, v3_txn)
      self.__insert_v3_entity(v3_entity, v3_txn)
      v3_resp = datastore_pb.CommitResponse()
      self.__make_v3_call('Commit', v3_txn, v3_resp)
      return v3_resp.cost.index_writes

    v3_get_req = datastore_pb.GetRequest()
    v3_get_req.transaction.CopyFrom(v3_txn)
    v3_get_req.key.append(v3_entity.key)
    v3_get_resp = datastore_pb.GetResponse()
    self.__make_v3_call('Get', v3_get_req, v3_get_resp)
    if v3_get_resp.entity[0].HasField('entity'):
      raise apiproxy_errors.ApplicationError(datastore_v4_pb2.Error.BAD_REQUEST,
                                             'Entity already exists.')
    v3_put_req = datastore_pb.PutRequest()
    v3_put_req.transaction.CopyFrom(v3_txn)
    v3_put_req.entity.append(v3_entity)
    v3_put_resp = datastore_pb.PutResponse()
    self.__make_v3_call('Put', v3_put_req, v3_put_resp)
    return v3_put_resp.cost.index_writes

  def __update_v3_entity(self, v3_entity, v3_txn):
    """Updates a v3 entity.

    Args:
      v3_entity: a datastore_v4_pb2.Entity
      v3_txn: a datastore_pb.Transaction or None

    Returns:
      the number of index writes that occurred

    Raises:
      ApplicationError: if the entity does not exist
    """
    if not v3_txn:

      v3_txn = datastore_pb.Transaction()
      v3_begin_txn_req = datastore_pb.BeginTransactionRequest()
      v3_begin_txn_req.app = v3_entity.key.app
      self.__make_v3_call('BeginTransaction', v3_begin_txn_req, v3_txn)
      self.__update_v3_entity(v3_entity, v3_txn)
      v3_resp = datastore_pb.CommitResponse()
      self.__make_v3_call('Commit', v3_txn, v3_resp)
      return v3_resp.cost.index_writes

    v3_get_req = datastore_pb.GetRequest()
    v3_get_req.transaction.CopyFrom(v3_txn)
    v3_get_req.key.append(v3_entity.key)
    v3_get_resp = datastore_pb.GetResponse()
    self.__make_v3_call('Get', v3_get_req, v3_get_resp)
    if not v3_get_resp.entity[0].HasField('entity'):
      raise apiproxy_errors.ApplicationError(datastore_v4_pb2.Error.BAD_REQUEST,
                                             'Entity does not exist.')
    v3_put_req = datastore_pb.PutRequest()
    v3_put_req.transaction.CopyFrom(v3_txn)
    v3_put_req.entity.append(v3_entity)
    v3_put_resp = datastore_pb.PutResponse()
    self.__make_v3_call('Put', v3_put_req, v3_put_resp)
    return v3_put_resp.cost.index_writes

  def __apply_v4_deprecated_mutation(self, v4_deprecated_mutation, v4_txn):
    """Applies a v4 DeprecatedMutation.

    Args:
      v4_deprecated_mutation: a datastore_v4_pb2.DeprecatedMutation
      v4_txn: an optional v4 transaction handle or None

    Returns:
      a datastore_v4_pb2.DeprecatedMutationResult
    """
    index_writes = 0
    v3_txn = None
    if v4_txn:
      v3_txn = datastore_pb.Transaction()
      self.__service_converter.v4_to_v3_txn(v4_txn, v3_txn)


    for v4_entity in v4_deprecated_mutation.insert:
      v3_entity = entity_pb2.EntityProto()
      self.__entity_converter.v4_to_v3_entity(v4_entity, v3_entity)
      index_writes += self.__insert_v3_entity(v3_entity, v3_txn)


    for v4_entity in v4_deprecated_mutation.update:
      v3_entity = entity_pb2.EntityProto()
      self.__entity_converter.v4_to_v3_entity(v4_entity, v3_entity)
      index_writes += self.__update_v3_entity(v3_entity, v3_txn)


    v3_insert_auto_req = datastore_pb.PutRequest()
    if v3_txn:
      v3_insert_auto_req.transaction.CopyFrom(v3_txn)
    for v4_entity in v4_deprecated_mutation.insert_auto_id:
      v3_entity = entity_pb2.EntityProto()
      self.__entity_converter.v4_to_v3_entity(v4_entity, v3_entity)
      v3_insert_auto_req.entity.append(v3_entity)
    v3_insert_auto_id_resp = datastore_pb.PutResponse()
    self.__make_v3_call('Put', v3_insert_auto_req, v3_insert_auto_id_resp)
    index_writes += v3_insert_auto_id_resp.cost.index_writes


    v3_upsert_req = datastore_pb.PutRequest()
    if v3_txn:
      v3_upsert_req.transaction.CopyFrom(v3_txn)
    for v4_entity in v4_deprecated_mutation.upsert:
      v3_entity = entity_pb2.EntityProto()
      self.__entity_converter.v4_to_v3_entity(v4_entity, v3_entity)
      v3_upsert_req.entity.append(v3_entity)
    v3_upsert_resp = datastore_pb.PutResponse()
    self.__make_v3_call('Put', v3_upsert_req, v3_upsert_resp)
    index_writes += v3_upsert_resp.cost.index_writes


    v3_delete_req = datastore_pb.DeleteRequest()
    if v3_txn:
      v3_delete_req.transaction.CopyFrom(v3_txn)
    for v4_key in v4_deprecated_mutation.delete:
      self.__entity_converter.v4_to_v3_reference(v4_key,
                                                 v3_delete_req.key.add())
    v3_delete_resp = datastore_pb.DeleteResponse()
    self.__make_v3_call('Delete', v3_delete_req, v3_delete_resp)
    index_writes += v3_delete_resp.cost.index_writes

    v4_deprecated_mutation_result = datastore_v4_pb2.DeprecatedMutationResult()
    for v3_ref in v3_insert_auto_id_resp.key:
      self.__entity_converter.v3_to_v4_key(
          v3_ref, v4_deprecated_mutation_result.insert_auto_id_key.add())
    v4_deprecated_mutation_result.index_updates = index_writes

    return v4_deprecated_mutation_result

  def __normalize_v4_run_query_request(self, v4_req):

    pass

  def __make_v3_call(self, method, v3_req, v3_resp):
    apiproxy_stub_map.MakeSyncCall(V3_SERVICE_NAME, method, v3_req, v3_resp)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()


from google.appengine.api import api_base_pb2 as google_dot_appengine_dot_api_dot_api__base__pb2
from google.appengine.api import mail_service_pb2 as google_dot_appengine_dot_api_dot_mail__service__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,google/appengine/api/mail_stub_service.proto\x12\x10google.appengine\x1a#google/appengine/api/api_base.proto\x1a\'google/appengine/api/mail_service.proto\"N\n\x17GetSentMessagesResponse\x12\x33\n\x0csent_message\x18\x01 \x03(\x0b\x32\x1d.google.appengine.MailMessage\"5\n\x19\x43learSentMessagesResponse\x12\x18\n\x10messages_cleared\x18\x01 \x01(\x05\"/\n\x16GetLogMailBodyResponse\x12\x15\n\rlog_mail_body\x18\x01 \x02(\x08\".\n\x15SetLogMailBodyRequest\x12\x15\n\rlog_mail_body\x18\x01 \x02(\x08\"1\n\x17GetLogMailLevelResponse\x12\x16\n\x0elog_mail_level\x18\x01 \x02(\t\"0\n\x16SetLogMailLevelRequest\x12\x16\n\x0elog_mail_level\x18\x01 \x02(\tB2\n\x1d\x63om.google.appengine.api.mailB\x11MailStubServicePb')



_GETSENTMESSAGESRESPONSE = DESCRIPTOR.message_types_by_name['GetSentMessagesResponse']
_CLEARSENTMESSAGESRESPONSE = DESCRIPTOR.message_types_by_name['ClearSentMessagesResponse']
_GETLOGMAILBODYRESPONSE = DESCRIPTOR.message_types_by_name['GetLogMailBodyResponse']
_SETLOGMAILBODYREQUEST = DESCRIPTOR.message_types_by_name['SetLogMailBodyRequest']
_GETLOGMAILLEVELRESPONSE = DESCRIPTOR.message_types_by_name['GetLogMailLevelResponse']
_SETLOGMAILLEVELREQUEST = DESCRIPTOR.message_types_by_name['SetLogMailLevelRequest']
GetSentMessagesResponse = _reflection.GeneratedProtocolMessageType('GetSentMessagesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSENTMESSAGESRESPONSE,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(GetSentMessagesResponse)

ClearSentMessagesResponse = _reflection.GeneratedProtocolMessageType('ClearSentMessagesResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLEARSENTMESSAGESRESPONSE,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(ClearSentMessagesResponse)

GetLogMailBodyResponse = _reflection.GeneratedProtocolMessageType('GetLogMailBodyResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETLOGMAILBODYRESPONSE,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(GetLogMailBodyResponse)

SetLogMailBodyRequest = _reflection.GeneratedProtocolMessageType('SetLogMailBodyRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETLOGMAILBODYREQUEST,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(SetLogMailBodyRequest)

GetLogMailLevelResponse = _reflection.GeneratedProtocolMessageType('GetLogMailLevelResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETLOGMAILLEVELRESPONSE,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(GetLogMailLevelResponse)

SetLogMailLevelRequest = _reflection.GeneratedProtocolMessageType('SetLogMailLevelRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETLOGMAILLEVELREQUEST,
  '__module__' : 'google.appengine.api.mail_stub_service_pb2'

  })
_sym_db.RegisterMessage(SetLogMailLevelRequest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.google.appengine.api.mailB\021MailStubServicePb'
  _GETSENTMESSAGESRESPONSE._serialized_start=144
  _GETSENTMESSAGESRESPONSE._serialized_end=222
  _CLEARSENTMESSAGESRESPONSE._serialized_start=224
  _CLEARSENTMESSAGESRESPONSE._serialized_end=277
  _GETLOGMAILBODYRESPONSE._serialized_start=279
  _GETLOGMAILBODYRESPONSE._serialized_end=326
  _SETLOGMAILBODYREQUEST._serialized_start=328
  _SETLOGMAILBODYREQUEST._serialized_end=374
  _GETLOGMAILLEVELRESPONSE._serialized_start=376
  _GETLOGMAILLEVELRESPONSE._serialized_end=425
  _SETLOGMAILLEVELREQUEST._serialized_start=427
  _SETLOGMAILLEVELREQUEST._serialized_end=475




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Stub version of the images API."""






import datetime
import json
import logging
import re
import time

import google

from google.appengine.api import apiproxy_stub
from google.appengine.api import apiproxy_stub_map
from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import images
from google.appengine.api.blobstore import blobstore_stub
from google.appengine.api.images import images_blob_stub
from google.appengine.api.images import images_service_pb2
from google.appengine.runtime import apiproxy_errors

from PIL import Image
import six

BLOB_SERVING_URL_KIND = images_blob_stub.BLOB_SERVING_URL_KIND
BMP = 'BMP'
GIF = 'GIF'
GS_INFO_KIND = '__GsFileInfo__'
ICO = 'ICO'
JPEG = 'JPEG'
MAX_REQUEST_SIZE = 32 << 20  # 32MB
PNG = 'PNG'
RGB = 'RGB'
RGBA = 'RGBA'
TIFF = 'TIFF'
WEBP = 'WEBP'

FORMAT_LIST = [BMP, GIF, ICO, JPEG, PNG, TIFF, WEBP]
EXIF_TIME_REGEX = re.compile(r'^([0-9]{4}):([0-9]{1,2}):([0-9]{1,2})'
                             ' ([0-9]{1,2}):([0-9]{1,2})(?::([0-9]{1,2}))?')


# Orientation tag id in EXIF.
_EXIF_ORIENTATION_TAG = 274

# DateTimeOriginal tag in EXIF.
_EXIF_DATETIMEORIGINAL_TAG = 36867

# Subset of EXIF tags. The stub is only able to extract these fields.
_EXIF_TAGS = {
    256: 'ImageWidth',
    257: 'ImageLength',
    271: 'Make',
    272: 'Model',
    _EXIF_ORIENTATION_TAG: 'Orientation',
    305: 'Software',
    306: 'DateTime',
    34855: 'ISOSpeedRatings',
    _EXIF_DATETIMEORIGINAL_TAG: 'DateTimeOriginal',
    36868: 'DateTimeDigitized',
    37383: 'MeteringMode',
    37385: 'Flash',
    41987: 'WhiteBalance'}


def _ArgbToRgbaTuple(argb):
  """Convert from a single ARGB value to a tuple containing RGBA.

  Args:
    argb: Signed 32 bit integer containing an ARGB value.

  Returns:
    RGBA tuple.
  """

  unsigned_argb = argb % 0x100000000
  return ((unsigned_argb >> 16) & 0xFF,
          (unsigned_argb >> 8) & 0xFF,
          unsigned_argb & 0xFF,
          (unsigned_argb >> 24) & 0xFF)


def _BackendPremultiplication(color):
  """Apply premultiplication and unpremultiplication to match production.

  Args:
    color: color tuple as returned by _ArgbToRgbaTuple.

  Returns:
    RGBA tuple.
  """




  alpha = color[3]
  rgb = color[0:3]
  multiplied = [(x * (alpha + 1)) >> 8 for x in rgb]
  if alpha:
    alpha_inverse = 0xffffff / alpha
    unmultiplied = [int(x * alpha_inverse) >> 16 for x in multiplied]
  else:
    unmultiplied = [0] * 3

  return tuple(unmultiplied + [alpha])


class ImagesServiceStub(apiproxy_stub.APIProxyStub):
  """Stub version of images API to be used with the dev_appserver."""


  THREADSAFE = False

  def __init__(self, service_name='images', host_prefix=''):
    """Preloads PIL to load all modules in the unhardened environment.

    Args:
      service_name: Service name expected for all calls.
      host_prefix: the URL prefix (protocol://host:port) to preprend to
        image urls on a call to GetUrlBase.
    """
    super(ImagesServiceStub, self).__init__(
        service_name, max_request_size=MAX_REQUEST_SIZE)
    self._blob_stub = images_blob_stub.ImagesBlobStub(host_prefix)
    Image.init()

  def _Dynamic_Composite(self, request, response):
    """Implementation of ImagesService::Composite.

    Based off documentation of the PIL library at
    http://www.pythonware.com/library/pil/handbook/index.htm

    Args:
      request: ImagesCompositeRequest - Contains image request info.
      response: ImagesCompositeResponse - Contains transformed image.

    Raises:
      ApplicationError: Bad data was provided, likely data about the dimensions.
    """
    if (not request.canvas.width or not request.canvas.height or
        not request.image or not request.options):
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
    if (request.canvas.width > 4000 or request.canvas.height > 4000 or
        len(request.options) > images.MAX_COMPOSITES_PER_REQUEST):
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

    width = request.canvas.width
    height = request.canvas.height
    color = _ArgbToRgbaTuple(request.canvas.color)


    color = _BackendPremultiplication(color)
    canvas = Image.new(RGBA, (width, height), color)
    sources = []
    for image in request.image:
      sources.append(self._OpenImageData(image))

    for options in request.options:
      if (options.anchor < images.TOP_LEFT or
          options.anchor > images.BOTTOM_RIGHT):
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
      if options.source_index >= len(sources) or options.source_index < 0:
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
      if options.opacity < 0 or options.opacity > 1:
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
      source = sources[options.source_index]
      x_anchor = (options.anchor % 3) * 0.5
      y_anchor = (options.anchor // 3) * 0.5
      x_offset = int(options.x_offset + x_anchor * (width - source.size[0]))
      y_offset = int(options.y_offset + y_anchor * (height - source.size[1]))
      if source.mode == RGBA:
        canvas.paste(source, (x_offset, y_offset), source)
      else:
        alpha = options.opacity * 255


        if six.PY3:
          alpha = int(round(alpha))
        mask = Image.new('L', source.size, alpha)
        canvas.paste(source, (x_offset, y_offset), mask)
    response_value = self._EncodeImage(canvas, request.canvas.output)
    response.image.content = response_value

  def _Dynamic_Histogram(self, request, response):
    """Trivial implementation of an API.

    Based off documentation of the PIL library at
    http://www.pythonware.com/library/pil/handbook/index.htm

    Args:
      request: ImagesHistogramRequest - Contains the image.
      response: ImagesHistogramResponse - Contains histogram of the image.

    Raises:
      ApplicationError: Image was of an unsupported format.
    """
    image = self._OpenImageData(request.image)

    img_format = image.format
    if img_format not in FORMAT_LIST:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.NOT_IMAGE)
    image = image.convert(RGBA)
    red = [0] * 256
    green = [0] * 256
    blue = [0] * 256




    for pixel in image.getdata():
      red[int((pixel[0] * pixel[3]) / 255)] += 1
      green[int((pixel[1] * pixel[3]) / 255)] += 1
      blue[int((pixel[2] * pixel[3]) / 255)] += 1
    histogram = response.histogram
    for value in red:
      histogram.red.append(value)
    for value in green:
      histogram.green.append(value)
    for value in blue:
      histogram.blue.append(value)

  def _Dynamic_Transform(self, request, response):
    """Trivial implementation of ImagesService::Transform.

    Based off documentation of the PIL library at
    http://www.pythonware.com/library/pil/handbook/index.htm

    Args:
      request: ImagesTransformRequest, contains image request info.
      response: ImagesTransformResponse, contains transformed image.
    """
    original_image = self._OpenImageData(request.image)

    input_settings = request.input
    correct_orientation = (
        input_settings.HasField('correct_exif_orientation') and
        input_settings.correct_exif_orientation ==
        images_service_pb2.InputSettings.CORRECT_ORIENTATION)



    source_metadata = self._ExtractMetadata(original_image,
                                            input_settings.parse_metadata)
    if input_settings.parse_metadata:
      logging.info(
          'Once the application is deployed, a more powerful metadata '
          'extraction will be performed which might return many more fields.')

    new_image = self._ProcessTransforms(original_image, request.transform,
                                        correct_orientation)

    substitution_rgb = None
    if input_settings.HasField('transparent_substitution_rgb'):
      substitution_rgb = input_settings.transparent_substitution_rgb
    response_value = self._EncodeImage(new_image, request.output,
                                       substitution_rgb)
    response.image.content = response_value
    response.source_metadata = source_metadata

  def _Dynamic_GetUrlBase(self, request, response):
    self._blob_stub.GetUrlBase(request, response)

  def _Dynamic_DeleteUrlBase(self, request, response):
    self._blob_stub.DeleteUrlBase(request, response)

  def _EncodeImage(self, image, output_encoding, substitution_rgb=None):
    """Encode the given image and return it in string form.

    Args:
      image: PIL Image object, image to encode.
      output_encoding: ImagesTransformRequest.OutputSettings object.
      substitution_rgb: The color to use for transparent pixels if the output
        format does not support transparency.

    Returns:
      str - Encoded image information in given encoding format.  Default is PNG.
    """
    image_string = six.BytesIO()
    image_encoding = PNG

    if output_encoding.mime_type == images_service_pb2.OutputSettings.WEBP:
      image_encoding = WEBP

    if output_encoding.mime_type == images_service_pb2.OutputSettings.JPEG:
      image_encoding = JPEG






      if substitution_rgb:



        blue = substitution_rgb & 0xFF
        green = (substitution_rgb >> 8) & 0xFF
        red = (substitution_rgb >> 16) & 0xFF
        background = Image.new(RGB, image.size, (red, green, blue))
        background.paste(image, mask=image.convert(RGBA).split()[3])
        image = background
      else:
        image = image.convert(RGB)

    image.save(image_string, image_encoding)
    return image_string.getvalue()

  def _OpenImageData(self, image_data):
    """Open image data from ImageData protocol buffer.

    Args:
      image_data: ImageData protocol buffer containing image data or blob
        reference.

    Returns:
      Image containing the image data passed in or reference by blob-key.

    Raises:
      ApplicationError: Both content and blob-key are provided.
      NOTE: 'content' must always be set because it is a required field,
      however, it must be the empty string when a blob-key is provided.
    """
    if image_data.content and image_data.HasField('blob_key'):
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.INVALID_BLOB_KEY)

    if image_data.HasField('blob_key'):
      image = self._OpenBlob(image_data.blob_key)
    else:
      image = self._OpenImage(image_data.content)


    img_format = image.format
    if img_format not in FORMAT_LIST:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.NOT_IMAGE)
    return image

  def _OpenImage(self, image):
    """Opens an image provided as a string.

    Args:
      image: Image data to be opened.

    Raises:
      ApplicationError: Image could not be opened or was an unsupported format.

    Returns:
      Image containing the image data passed in.
    """
    if not image:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.NOT_IMAGE)

    image = six.BytesIO(image)
    try:
      return Image.open(image)
    except IOError:

      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_IMAGE_DATA)

  def _OpenBlob(self, blob_key):
    """Create an Image from the blob data read from blob_key."""

    try:
      _ = datastore.Get(
          blobstore_stub.BlobstoreServiceStub.ToDatastoreBlobKey(blob_key))
    except datastore_errors.Error:


      logging.exception('Blob with key %r does not exist', blob_key)
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.UNSPECIFIED_ERROR)

    blobstore_storage = apiproxy_stub_map.apiproxy.GetStub('blobstore')


    try:
      blob_file = blobstore_storage.storage.OpenBlob(blob_key)
    except IOError:
      logging.exception('Could not get file for blob_key %r', blob_key)

      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_IMAGE_DATA)

    try:
      return Image.open(blob_file)
    except IOError:
      logging.exception('Could not open image %r for blob_key %r',
                        blob_file, blob_key)

      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_IMAGE_DATA)

  def _ValidateCropArg(self, arg):
    """Check an argument for the Crop transform.

    Args:
      arg: float - Argument to Crop transform to check.

    Raises:
      ApplicationError: There was a problem with the provided argument.
    """
    if not isinstance(arg, float):
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

    if 0 > arg or arg > 1.0:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

  def _CalculateNewDimensions(self,
                              current_width,
                              current_height,
                              req_width,
                              req_height,
                              crop_to_fit,
                              allow_stretch):
    """Get new resize dimensions keeping the current aspect ratio.

    This uses the more restricting of the two requested values to determine
    the new ratio. See also crop_to_fit.

    Args:
      current_width: int, current width of the image.
      current_height: int, current height of the image.
      req_width: int, requested new width of the image, 0 if unspecified.
      req_height: int, requested new height of the image, 0 if unspecified.
      crop_to_fit: bool, True if the less restricting dimension should be used.
      allow_stretch: bool, True is aspect ratio should be ignored.

    Raises:
      apiproxy_errors.ApplicationError: if crop_to_fit is True either req_width
        or req_height is 0.

    Returns:
      tuple (width, height) ints of the new dimensions.
    """


    width_ratio = float(req_width) / current_width
    height_ratio = float(req_height) / current_height

    height = req_height
    width = req_width
    if allow_stretch or crop_to_fit:

      if not req_width or not req_height:
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
      if not allow_stretch:
        if width_ratio > height_ratio:
          height = int(width_ratio * current_height)
        else:
          width = int(height_ratio * current_width)
    else:



      if not req_width or (width_ratio > height_ratio and req_height):

        width = int(height_ratio * current_width)
      else:

        height = int(width_ratio * current_height)
    return width, height

  def _Resize(self, image, transform):
    """Use PIL to resize the given image with the given transform.

    Args:
      image: PIL.Image.Image object to resize.
      transform: images_service_pb2.Transform to use when resizing.

    Returns:
      PIL.Image.Image with transforms performed on it.

    Raises:
      ApplicationError: The resize data given was bad.
    """
    width = 0
    height = 0

    if transform.HasField('width'):
      width = transform.width
      if width < 0 or 4000 < width:
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

    if transform.HasField('height'):
      height = transform.height
      if height < 0 or 4000 < height:
        raise apiproxy_errors.ApplicationError(
            images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

    crop_to_fit = transform.crop_to_fit
    allow_stretch = transform.allow_stretch

    current_width, current_height = image.size
    new_width, new_height = self._CalculateNewDimensions(
        current_width, current_height, width, height, crop_to_fit,
        allow_stretch)
    new_image = image.resize((new_width, new_height), Image.ANTIALIAS)
    if crop_to_fit and (new_width > width or new_height > height):

      left = int((new_width - width) * transform.crop_offset_x)
      top = int((new_height - height) * transform.crop_offset_y)
      right = left + width
      bottom = top + height
      new_image = new_image.crop((left, top, right, bottom))

    return new_image

  def _Rotate(self, image, transform):
    """Use PIL to rotate the given image with the given transform.

    Args:
      image: PIL.Image.Image object to rotate.
      transform: images_service_pb2.Transform to use when rotating.

    Returns:
      PIL.Image.Image with transforms performed on it.

    Raises:
      ApplicationError: Given data for the rotate was bad.
    """
    degrees = transform.rotate
    if degrees < 0 or degrees % 90 != 0:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)
    degrees %= 360


    degrees = 360 - degrees
    return image.rotate(degrees)

  def _Crop(self, image, transform):
    """Use PIL to crop the given image with the given transform.

    Args:
      image: PIL.Image.Image object to crop.
      transform: images_service_pb2.Transform to use when cropping.

    Returns:
      PIL.Image.Image with transforms performed on it.

    Raises:
      BadRequestError if the crop data given is bad.
    """
    left_x = 0.0
    top_y = 0.0
    right_x = 1.0
    bottom_y = 1.0

    if transform.HasField('crop_left_x'):
      left_x = transform.crop_left_x
      self._ValidateCropArg(left_x)

    if transform.HasField('crop_top_y'):
      top_y = transform.crop_top_y
      self._ValidateCropArg(top_y)

    if transform.HasField('crop_right_x'):
      right_x = transform.crop_right_x
      self._ValidateCropArg(right_x)

    if transform.HasField('crop_bottom_y'):
      bottom_y = transform.crop_bottom_y
      self._ValidateCropArg(bottom_y)


    width, height = image.size

    box = (int(round(left_x * width)),
           int(round(top_y * height)),
           int(round(right_x * width)),
           int(round(bottom_y * height)))

    return image.crop(box)

  @staticmethod
  def _GetExifFromImage(image):
    if hasattr(image, '_getexif'):














      try:

        from PIL import TiffImagePlugin

        return image._getexif()
      except ImportError:
        # We have not managed to get this to work in the SDK with Python
        # 2.5, so just catch the ImportError and pretend there is no
        # EXIF information of interest.
        logging.info(
            'Sorry, TiffImagePlugin does not work in this environment')

    return None

  @staticmethod
  def _ExtractMetadata(image, parse_metadata):
    """Extract EXIF metadata from the image.

    Note that this is a much simplified version of metadata extraction. After
    deployment applications have access to a more powerful parser that can
    parse hundreds of fields from images.

    Args:
      image: PIL Image object.
      parse_metadata: bool, True if metadata parsing has been requested. If
        False the result will contain image dimensions.
    Returns:
      str - JSON encoded values with various metadata fields.
    """

    def ExifTimeToUnixtime(exif_time):
      """Convert time in EXIF to unix time.

      Args:
        exif_time: str - Time from the EXIF block formatted by EXIF standard.
            Seconds are optional.  (Example: '2011:02:20 10:23:12')

      Returns:
        Integer, the time in unix fromat: seconds since the epoch.
      """
      match = EXIF_TIME_REGEX.match(exif_time)
      if not match:
        return None
      try:
        pieces = [int(g) for g in match.groups() if g]
        date = datetime.datetime(*pieces)
      except ValueError:
        logging.info('Invalid date in EXIF: %s', exif_time)
        return None
      return int(time.mktime(date.timetuple()))

    metadata_dict = (
        parse_metadata and ImagesServiceStub._GetExifFromImage(image) or {})

    metadata_dict[256], metadata_dict[257] = image.size



    if _EXIF_DATETIMEORIGINAL_TAG in metadata_dict:
      date_ms = ExifTimeToUnixtime(metadata_dict[_EXIF_DATETIMEORIGINAL_TAG])
      if date_ms:
        metadata_dict[_EXIF_DATETIMEORIGINAL_TAG] = date_ms
      else:
        del metadata_dict[_EXIF_DATETIMEORIGINAL_TAG]
    metadata = dict([(_EXIF_TAGS[k], v)
                     for k, v in six.iteritems(metadata_dict)
                     if k in _EXIF_TAGS])
    return json.dumps(metadata)

  def _CorrectOrientation(self, image, orientation):
    """Use PIL to correct the image orientation based on its EXIF.

    See JEITA CP-3451 at http://www.exif.org/specifications.html,
    Exif 2.2, page 18.

    Args:
      image: source PIL.Image.Image object.
      orientation: integer in range (1,8) inclusive, corresponding the image
        orientation from EXIF.

    Returns:
      PIL.Image.Image with transforms performed on it. If no correction was
        done, it returns the input image.
    """


    if orientation == 2:
      image = image.transpose(Image.FLIP_LEFT_RIGHT)
    elif orientation == 3:
      image = image.rotate(180)
    elif orientation == 4:
      image = image.transpose(Image.FLIP_TOP_BOTTOM)
    elif orientation == 5:
      image = image.transpose(Image.FLIP_TOP_BOTTOM)
      image = image.rotate(270)
    elif orientation == 6:
      image = image.rotate(270)
    elif orientation == 7:
      image = image.transpose(Image.FLIP_LEFT_RIGHT)
      image = image.rotate(270)
    elif orientation == 8:
      image = image.rotate(90)

    return image

  def _ProcessTransforms(self, image, transforms, correct_orientation):
    """Execute PIL operations based on transform values.

    Args:
      image: PIL.Image.Image instance, image to manipulate.
      transforms: list of ImagesTransformRequest.Transform objects.
      correct_orientation: True to indicate that image orientation should be
        corrected based on its EXIF.
    Returns:
      PIL.Image.Image with transforms performed on it.

    Raises:
      ApplicationError: More than one of the same type of transform was present.
    """
    new_image = image
    if len(transforms) > images.MAX_TRANSFORMS_PER_REQUEST:
      raise apiproxy_errors.ApplicationError(
          images_service_pb2.ImagesServiceError.BAD_TRANSFORM_DATA)

    orientation = 1
    if correct_orientation:


      exif = self._GetExifFromImage(image)
      if not exif or _EXIF_ORIENTATION_TAG not in exif:
        correct_orientation = False
      else:
        orientation = exif[_EXIF_ORIENTATION_TAG]

      width, height = new_image.size
      if height > width:
        orientation = 1

    for transform in transforms:







      if (correct_orientation and not (transform.HasField('crop_left_x') or
                                       transform.HasField('crop_top_y') or
                                       transform.HasField('crop_right_x') or
                                       transform.HasField('crop_bottom_y')) and
          not transform.HasField('horizontal_flip') and
          not transform.HasField('vertical_flip')):
        new_image = self._CorrectOrientation(new_image, orientation)
        correct_orientation = False

      if transform.HasField('width') or transform.HasField('height'):

        new_image = self._Resize(new_image, transform)

      elif transform.HasField('rotate'):

        new_image = self._Rotate(new_image, transform)

      elif transform.HasField('horizontal_flip'):

        new_image = new_image.transpose(Image.FLIP_LEFT_RIGHT)

      elif transform.HasField('vertical_flip'):

        new_image = new_image.transpose(Image.FLIP_TOP_BOTTOM)

      elif (transform.HasField('crop_left_x') or
            transform.HasField('crop_top_y') or
            transform.HasField('crop_right_x') or
            transform.HasField('crop_bottom_y')):

        new_image = self._Crop(new_image, transform)

      elif transform.HasField('autolevels'):


        logging.info('I\'m Feeling Lucky autolevels will be visible once this '
                     'application is deployed.')
      else:
        logging.warning('Found no transformations found to perform.')

      if correct_orientation:


        new_image = self._CorrectOrientation(new_image, orientation)
        correct_orientation = False




    return new_image



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""URL downloading API."""

import email
import threading

import six
from six.moves import urllib
from six.moves import UserDict as IterableUserDict
import six.moves.http_client
import six.moves.urllib.parse

from google.appengine.api import apiproxy_stub_map
from google.appengine.api import urlfetch_service_pb2
from google.appengine.api.urlfetch_errors import *
from google.appengine.runtime import apiproxy_errors
from google.appengine.runtime import context












MAX_REDIRECTS = 5


GET = 1
POST = 2
HEAD = 3
PUT = 4
DELETE = 5
PATCH = 6

_URL_STRING_MAP = {
    'GET': GET,
    'POST': POST,
    'HEAD': HEAD,
    'PUT': PUT,
    'DELETE': DELETE,
    'PATCH': PATCH,
}

_VALID_METHODS = frozenset(list(_URL_STRING_MAP.values()))

_thread_local_settings = threading.local()


class _CaselessDict(IterableUserDict):
  """Case insensitive `dict`.

  This class was lifted from os.py and slightly modified.
  """

  def __init__(self, dict=None, **kwargs):
    self.caseless_keys = {}
    IterableUserDict.__init__(self, dict, **kwargs)

  def __setitem__(self, key, item):
    """Sets `dict` item.

    Args:
      key: Key of new item. Key is case insensitive, so `d['Key'] = value` will
        replace previous values set by `d['key'] = old_value`.
      item: Item to store.
    """
    caseless_key = key.lower()

    if caseless_key in self.caseless_keys:
      del self.data[self.caseless_keys[caseless_key]]
    self.caseless_keys[caseless_key] = key
    self.data[key] = item

  def __getitem__(self, key):
    """Gets `dict` item.

    Args:
      key: Key of item to get. Key is case insensitive, so `d['Key']` is the
        same as `d['key']`.

    Returns:
      Item associated with `key`.

    Raises:
      KeyError: If the `key` is not found.
    """
    return self.data[self.caseless_keys[key.lower()]]

  def __delitem__(self, key):
    """Removes item from `dict`.

    Args:
      key: Key of item to remove.  Key is case insensitive, so `del d['Key']` is
        the same as `del d['key']`.
    """
    caseless_key = key.lower()
    del self.data[self.caseless_keys[caseless_key]]
    del self.caseless_keys[caseless_key]

  def has_key(self, key):
    """Determines if the `dict` has an item with a specific `key`.

    Args:
      key: Key to check for presence. Key is case insensitive, so
        `d.has_key('Key')` evaluates to the same value as `d.has_key('key')`.

    Returns:
      `True` if `dict` contains the specified `key`, else `False`.
    """
    return key.lower() in self.caseless_keys

  def __contains__(self, key):
    """Same as `has_key`, but used for `in` operator."""
    return self.has_key(key)

  def get(self, key, failobj=None):
    """Gets `dict` item, defaulting to another value if it does not exist.

    Args:
      key: Key of item to get. Key is case insensitive, so `d['Key']` is the
        same as `d['key']`.
      failobj: Value to return if key not in `dict`.

    Returns:
      A `dict` item.
    """
    try:
      cased_key = self.caseless_keys[key.lower()]
    except KeyError:
      return failobj
    return self.data[cased_key]

  def update(self, dict=None, **kwargs):
    """Updates the `dict` using values from another `dict` and keywords.

    Args:
      dict: Dictionary to update from.
      **kwargs: Keyword arguments to update from.
    """
    if dict:
      try:
        keys = list(dict.keys())
      except AttributeError:

        for k, v in dict:
          self[k] = v
      else:



        for k in keys:
          self[k] = dict[k]
    if kwargs:
      self.update(kwargs)

  def copy(self):
    """Makes a shallow, case-sensitive copy of `self`.

    Returns:
      A `dict` copy of `self`.
    """
    return dict(self)


def _is_fetching_self(url, method):
  """Checks if the fetch is for the same URL from which it originated.

  Args:
    url: `string`. The URL being fetched.
    method: Value from `_VALID_METHODS`.

  Returns:
    Boolean indicating whether or not it seems that the app is trying to fetch
        itself.
  """
  if (method != GET or context.get('HTTP_HOST', None) is None or
      context.get('PATH_INFO', None) is None):
    return False

  _, host_port, path, _, _ = six.moves.urllib.parse.urlsplit(url)

  if host_port == context.get('HTTP_HOST'):

    current_path = urllib.parse.unquote(context.get('PATH_INFO'))
    desired_path = urllib.parse.unquote(path)

    if (current_path == desired_path or
        (current_path in ('', '/') and desired_path in ('', '/'))):
      return True

  return False


def create_rpc(deadline=None, callback=None):
  """Creates an RPC object for use with the URL Fetch API.

  Args:
    deadline: Optional deadline in seconds for the operation; the default is a
      system-specific deadline (typically 5 seconds).
    callback: Optional callable to invoke on completion.

  Returns:
    An `apiproxy_stub_map.UserRPC` object specialized for this service.
  """
  if deadline is None:
    deadline = get_default_fetch_deadline()
  return apiproxy_stub_map.UserRPC('urlfetch', deadline, callback)


def fetch(url, payload=None, method=GET, headers={},
          allow_truncated=False, follow_redirects=True,
          deadline=None, validate_certificate=None):
  """Fetches the given HTTP `url`, blocking until the result is returned.

  URLs are fetched using one of the following HTTP methods:
      - `GET`
      - `POST`
      - `HEAD`
      - `PUT`
      - `DELETE`
      - `PATCH`

  To fetch the result, a HTTP/1.1-compliant proxy is used.

  Args:
    method: The constants `GET`, `POST`, `HEAD`, `PUT`, `DELETE`, or `PATCH` or
      the same `HTTP` methods as strings.
    payload: `POST`, `PUT`, or `PATCH` payload (implies method is not `GET`,
      `HEAD`, or `DELETE`). This argument is ignored if the method is not
      `POST`, `PUT`, or `PATCH`.
    headers: Dictionary of `HTTP` headers to send with the request.
    allow_truncated: If set to `True`, truncates large responses and returns
      them without raising an error. Otherwise, a `ResponseTooLargeError` is
      raised when a response is truncated.
    follow_redirects: If set to `True` (the default), redirects are
      transparently followed. If there are less than five redirects, the
      response contains the final destination's payload, and the response status
      is `200`. You lose, however, the redirect chain information. If set to
      `False`, you see the `HTTP` response yourself, including the `Location`
      header, and redirects are not followed.
    deadline: Deadline in seconds for the operation.
    validate_certificate: If set to `True`, requests are not sent to the server
      unless the certificate is valid, signed by a trusted `CA`, and the host
      name matches the certificate. A value of `None` (default) indicates that
      the behavior will be chosen by the underlying `urlfetch` implementation.

  Returns:
    object: An object containing following fields:

        - `content`: A `string` that contains the response from the server.
        - `status_code`: The `HTTP` status code that was returned by the server.
        - `headers`: The `dict` of headers that was returned by the server.

  Raises:
    urlfetch_errors.Error: If an error occurs. See the `urlfetch_errors` module
        for more information.


  Note:
      `HTTP` errors are returned as a part of the return structure. `HTTP`
      errors like `404` do not result in an exception. See `urlfetch_errors` for
      details.

  """

  payload = six.ensure_binary(payload, 'utf-8') if payload else None
  rpc = create_rpc(deadline=deadline)
  make_fetch_call(rpc, url, payload, method, headers,
                  allow_truncated, follow_redirects, validate_certificate)
  return rpc.get_result()


def make_fetch_call(rpc, url, payload=None, method=GET, headers={},
                    allow_truncated=False, follow_redirects=True,
                    validate_certificate=None):
  """Executes the RPC call to fetch a given `HTTP` URL.

  The first argument is a `UserRPC` instance.  See `urlfetch.fetch` for a
  thorough description of the remaining arguments.

  Raises:
    InvalidMethodError: If the requested method is not in `_VALID_METHODS`.
    ResponseTooLargeError: If the response payload is too large.
    InvalidURLError: If there are issues with the content or size of the
        requested `url`.

  Returns:
    The `rpc` object that was passed into the function.

  """

  assert rpc.service == 'urlfetch', repr(rpc.service)
  if isinstance(method, six.string_types):
    method = method.upper()
  method = _URL_STRING_MAP.get(method, method)
  if method not in _VALID_METHODS:
    raise InvalidMethodError('Invalid method %s.' % str(method))

  if _is_fetching_self(url, method):
    raise InvalidURLError('App cannot fetch '
                          'the same `url` as the one used for the request.')

  request = urlfetch_service_pb2.URLFetchRequest()
  response = urlfetch_service_pb2.URLFetchResponse()

  request.Url = url

  if method == GET:
    request.Method = urlfetch_service_pb2.URLFetchRequest.GET
  elif method == POST:
    request.Method = urlfetch_service_pb2.URLFetchRequest.POST
  elif method == HEAD:
    request.Method = urlfetch_service_pb2.URLFetchRequest.HEAD
  elif method == PUT:
    request.Method = urlfetch_service_pb2.URLFetchRequest.PUT
  elif method == DELETE:
    request.Method = urlfetch_service_pb2.URLFetchRequest.DELETE
  elif method == PATCH:
    request.Method = urlfetch_service_pb2.URLFetchRequest.PATCH


  if payload and method in (POST, PUT, PATCH):

    request.Payload = six.ensure_binary(payload, 'utf-8')


  for key, value in six.iteritems(headers):




    request.header.add(Key=key, Value=str(value))

  request.FollowRedirects = follow_redirects
  if validate_certificate is not None:
    request.MustValidateServerCertificate = validate_certificate

  if rpc.deadline is not None:
    request.Deadline = rpc.deadline



  rpc.make_call('Fetch', request, response, _get_fetch_result, allow_truncated)
  return rpc


def _get_fetch_result(rpc):
  """Checks for success, handles exceptions, and returns a converted RPC result.

  This method waits for the `rpc` if it has not yet finished and calls the
  post-call hooks on the first invocation.

  Args:
    rpc: A `UserRPC` object.

  Raises:
    InvalidURLError: If the URL was invalid.
    DownloadError: If there was a problem fetching the URL.
    PayloadTooLargeError: If the request and its payload was larger than the
        allowed limit.
    ResponseTooLargeError: If the response was either truncated (and
        `allow_truncated=False` was passed to `make_fetch_call()`), or if it
        was too big for us to download.
    MalformedReplyError: If an invalid `HTTP` response was returned.
    TooManyRedirectsError: If the redirect limit was hit while `follow_rediects`
        was set to `True`.
    InternalTransientError: An internal error occurred. Wait a few minutes, then
        try again.
    ConnectionClosedError: If the target server prematurely closed the
        connection.
    DNSLookupFailedError: If the `DNS` lookup for the URL failed.
    DeadlineExceededError: If the deadline was exceeded; occurs when the
        client-supplied `deadline` is invalid or if the client did not specify a
        `deadline` and the system default value is invalid.
    SSLCertificateError: If an invalid server certificate was presented.
    AssertionError: If the `assert` statement fails.

  Returns:
    A `_URLFetchResult` object.
  """
  assert rpc.service == 'urlfetch', repr(rpc.service)
  assert rpc.method == 'Fetch', repr(rpc.method)

  url = rpc.request.Url

  try:
    rpc.check_success()
  except apiproxy_errors.RequestTooLargeError as err:
    raise InvalidURLError(
        'Request body too large fetching URL: ' + url)
  except apiproxy_errors.ApplicationError as err:
    error_detail = ''
    if err.error_detail:
      error_detail = ' Error: ' + err.error_detail
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.INVALID_URL):
      raise InvalidURLError(
          'Invalid request URL: ' + url + error_detail)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.PAYLOAD_TOO_LARGE):

      raise PayloadTooLargeError(
          'Request exceeds 10 MiB limit for URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.CLOSED):
      raise ConnectionClosedError(
          'Connection closed unexpectedly by server at URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.TOO_MANY_REDIRECTS):
      raise TooManyRedirectsError(
          'Too many redirects at URL: ' + url + ' with redirect=true')
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.MALFORMED_REPLY):
      raise MalformedReplyError(
          'Malformed HTTP reply received from server at URL: '
          + url + error_detail)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.INTERNAL_TRANSIENT_ERROR):
      raise InternalTransientError(
          'Temporary error in fetching URL: ' + url + ', please re-try')
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.DNS_ERROR):
      raise DNSLookupFailedError('DNS lookup failed for URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.UNSPECIFIED_ERROR):
      raise DownloadError('Unspecified error in fetching URL: '
                          + url + error_detail)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.FETCH_ERROR):
      raise DownloadError("Unable to fetch URL: " + url + error_detail)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.RESPONSE_TOO_LARGE):
      raise ResponseTooLargeError('HTTP response too large from URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.DEADLINE_EXCEEDED):
      raise DeadlineExceededError(
          'Deadline exceeded while waiting for HTTP response from URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.SSL_CERTIFICATE_ERROR):
      raise SSLCertificateError(
          'Invalid and/or missing SSL certificate for URL: ' + url)
    if (err.application_error ==
        urlfetch_service_pb2.URLFetchServiceError.CONNECTION_ERROR):
      raise DownloadError('Unable to connect to server at URL: ' + url)

    raise err

  response = rpc.response
  allow_truncated = rpc.user_data
  result = _URLFetchResult(response)
  if response.ContentWasTruncated and not allow_truncated:
    raise ResponseTooLargeError(result)
  return result


Fetch = fetch


def CreateHTTPHeaders(data):
  """Creates `HTTPHeaders` that are compatible with both Python 2 or Python 3."""
  if six.PY2:
    return six.moves.http_client.HTTPMessage(six.StringIO(data))
  else:
    return email.parser.HeaderParser(HTTPMessageWrapper).parsestr(data)




class HTTPMessageWrapper(six.moves.http_client.HTTPMessage):

  def getheaders(self, key):
    return self.get_all(key)

  @property
  def type(self):
    return self.get('content-type').split(';')[0]


class _URLFetchResult(object):
  """A Pythonic representation of our fetch response protocol buffer."""

  def __init__(self, response_proto):
    """Constructor.

    Args:
      response_proto: The `URLFetchResponse` protocol buffer to wrap.
    """
    self.__pb = response_proto
    self.content = response_proto.Content
    self.status_code = response_proto.StatusCode
    self.content_was_truncated = response_proto.ContentWasTruncated
    self.final_url = response_proto.FinalUrl or None

    data = ''.join(
        ['%s: %s\n' % (h.Key, h.Value) for h in response_proto.header] + ['\n'])
    self.header_msg = CreateHTTPHeaders(data)




    header_keys_values = [(key, self.header_msg.getheaders(key))
                          for key in self.header_msg.keys()]
    header_list = [
        (key, ', '.join(values)) for (key, values) in header_keys_values
    ]
    self.headers = _CaselessDict(header_list)


def get_default_fetch_deadline():
  """Gets the default value for `create_rpc()`'s `deadline` parameter."""
  return getattr(_thread_local_settings, "default_fetch_deadline", None)


def set_default_fetch_deadline(value):
  """Sets the default value for `create_rpc()`'s `deadline` parameter.

  This setting is thread-specific, meaning it that is stored in a thread local.
  This function doesn't check the type or range of the value.  The default
  value is `None`.

  See also: `create_rpc()`, `fetch()`

  Args:
    value: The default value that you want to use for the `deadline` parameter
        of `create_rpc()`.

  """
  _thread_local_settings.default_fetch_deadline = value



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#














"""Low-level utilities used internally by NDB.

These are not meant for use by code outside NDB.
"""

import functools
import logging
import os
import sys

import six

__all__ = []

DEBUG = True



def logging_debug(*args):


  if DEBUG and logging.getLogger().level < logging.DEBUG:
    logging.debug(*args)


def _wrapping(wrapped):
  """A decorator to decorate a decorator's wrapper.

  Following the lead of Twisted and Monocle, this is supposed to make debugging
  heavily decorated code easier.  We'll see...
  # TODO: This copies the functionality of functools.wraps
  # following the patch in http://bugs.python.org/issue3445. We can replace
  # this once upgrading to python 3.3.

  Args:
    wrapped: The decorator to wrap.

  Returns:
    The decorator, now wrapped.
  """
  def wrapping_wrapper(wrapper):
    try:
      wrapper.__wrapped__ = wrapped
      wrapper.__name__ = wrapped.__name__
      wrapper.__doc__ = wrapped.__doc__
      wrapper.__dict__.update(wrapped.__dict__)

      if hasattr(wrapped, '__module__'):
        wrapper.__module__ = wrapped.__module__
    except Exception:
      pass
    return wrapper
  return wrapping_wrapper








wrapping = _wrapping if six.PY2 else functools.wraps


def get_stack(limit=10):

  if not DEBUG:
    return ()
  frame = sys._getframe(1)
  lines = []
  while len(lines) < limit and frame is not None:
    f_locals = frame.f_locals
    ndb_debug = f_locals.get('__ndb_debug__')
    if ndb_debug != 'SKIP':
      line = frame_info(frame)
      if ndb_debug is not None:
        line += ' # ' + str(ndb_debug)
      lines.append(line)
    frame = frame.f_back
  return lines


def func_info(func, lineno=None):
  if not DEBUG:
    return None
  func = getattr(func, '__wrapped__', func)
  code = getattr(func, 'func_code' if six.PY2 else '__code__', None)
  return code_info(code, lineno)


def gen_info(gen):
  if not DEBUG:
    return None
  frame = gen.gi_frame
  if gen.gi_running:
    prefix = 'running generator '
  elif frame:
    if frame.f_lasti < 0:
      prefix = 'initial generator '
    else:
      prefix = 'suspended generator '
  else:
    prefix = 'terminated generator '
  if frame:
    return prefix + frame_info(frame)
  code = getattr(gen, 'gi_code', None)
  if code:
    return prefix + code_info(code)
  return prefix + hex(id(gen))


def frame_info(frame):
  if not DEBUG:
    return None
  return code_info(frame.f_code, frame.f_lineno)


def code_info(code, lineno=None):
  if not DEBUG or not code:
    return ''
  funcname = code.co_name


  filename = os.path.basename(code.co_filename)
  if lineno is None:
    lineno = code.co_firstlineno
  return '%s(%s:%s)' % (funcname, filename, lineno)


def positional(max_pos_args):
  """A decorator to declare that only the first N arguments may be positional.

  Note that for methods, n includes 'self'.
  """
  __ndb_debug__ = 'SKIP'

  def positional_decorator(wrapped):
    if not DEBUG:
      return wrapped
    __ndb_debug__ = 'SKIP'

    @wrapping(wrapped)
    def positional_wrapper(*args, **kwds):
      __ndb_debug__ = 'SKIP'
      if len(args) > max_pos_args:
        plural_s = ''
        if max_pos_args != 1:
          plural_s = 's'
        raise TypeError(
            '%s() takes at most %d positional argument%s (%d given)' %
            (wrapped.__name__, max_pos_args, plural_s, len(args)))
      return wrapped(*args, **kwds)
    return positional_wrapper
  return positional_decorator


def decorator(wrapped_decorator):
  """Converts a function into a decorator that optionally accepts keyword
  arguments in its declaration.

  Example usage:
    @utils.decorator
    def decorator(func, args, kwds, op1=None):
      ... apply op1 ...
      return func(*args, **kwds)

    # Form (1), vanilla
    @decorator
    foo(...)
      ...

    # Form (2), with options
    @decorator(op1=5)
    foo(...)
      ...

  Args:
    wrapped_decorator: A function that accepts positional args (func, args,
      kwds) and any additional supported keyword arguments.

  Returns:
    A decorator with an additional 'wrapped_decorator' property that is set to
  the original function.
  """
  def helper(_func=None, **options):
    def outer_wrapper(func):
      @wrapping(func)
      def inner_wrapper(*args, **kwds):
        return wrapped_decorator(func, args, kwds, **options)
      return inner_wrapper

    if _func is None:

      return outer_wrapper


    if options:

      raise TypeError('positional arguments not supported')
    return outer_wrapper(_func)
  helper.wrapped_decorator = wrapped_decorator
  return helper


def tweak_logging():






  q = 0
  v = 0
  for arg in sys.argv[1:]:
    if arg.startswith('-v'):
      v += arg.count('v')
    if arg.startswith('-q'):
      q += arg.count('q')
  if v >= 2:
    level = logging.INFO
    if v >= 3:
      level = logging.DEBUG - 1
    logging.basicConfig(level=level)
  if q > 0:
    global DEBUG
    DEBUG = False


if 'test' in os.path.basename(sys.argv[0]):
  tweak_logging()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Validation tools for generic object structures.

This library is used for defining classes with constrained attributes.
Attributes are defined on the class which contains them using validators.
Although validators can be defined by any client of this library, a number
of standard validators are provided here.

Validators can be any callable that takes a single parameter which checks
the new value before it is assigned to the attribute. Validators are
permitted to modify a received value so that it is appropriate for the
attribute definition. For example, using int as a validator will cast
a correctly formatted string to a number, or raise an exception if it
can not. This is not recommended, however. The correct way to use a
validator that ensures the correct type is to use the `Type` validator.

This validation library is mainly intended for use with the YAML object
builder. See the `yaml_object` module.
"""






import re
import google
from ruamel import yaml
import six


class SortedDict(dict):
  """Represents a dict with a particular key order for yaml representing."""

  def __init__(self, keys, data):
    super(SortedDict, self).__init__()
    self.keys = keys
    self.update(data)

  def ordered_items(self):
    result = []
    for key in self.keys:
      if self.get(key) is not None:
        result.append((key, self.get(key)))
    return result


class ItemDumper(yaml.SafeDumper):
  """For dumping `validation.Items`. Respects `SortedDict` key ordering."""

  def represent_mapping(self, tag, mapping, flow_style=None):
    if hasattr(mapping, 'ordered_items'):
      return super(ItemDumper, self).represent_mapping(
          tag, mapping.ordered_items(), flow_style=flow_style)
    return super(ItemDumper, self).represent_mapping(
        tag, mapping, flow_style=flow_style)


ItemDumper.add_representer(
    SortedDict, ItemDumper.represent_dict)


class Error(Exception):
  """Base class for all package errors."""


class AttributeDefinitionError(Error):
  """An error occurred in the definition of class attributes."""


class ValidationError(Error):
  """Base class for raising exceptions during validation."""

  def __init__(self, message, cause=None):
    """Initialize exception."""
    if hasattr(cause, 'args') and cause.args:
      Error.__init__(self, message, *cause.args)
    else:


      Error.__init__(self, message)
    self.message = message
    self.cause = cause

  def __str__(self):
    return str(self.message)


class MissingAttribute(ValidationError):
  """Raised when a required attribute is missing from object."""

  def __init__(self, key):
    msg = 'Missing required value [{}].'.format(key)
    super(MissingAttribute, self).__init__(msg)


def AsValidator(validator):
  """Wrap various types as instances of a validator.

  Used to allow shorthand for common validator types.  It
  converts the following types to the following Validators:

  *  strings -> Regex
  *  type -> Type
  *  collection -> Options
  *  Validator -> Itself

  Args:
    validator: Object to wrap in a validator.

  Returns:
    Validator instance that wraps the given value.

  Raises:
    AttributeDefinitionError: If `validator` is not one of the above described
      types.
  """
  if (_IsPy2Basestring(validator)
      or validator == six.string_types):
    return StringValidator()
  if isinstance(validator, (str, six.text_type)):
    return Regex(validator, type(validator))
  if isinstance(validator, type):
    return Type(validator)
  if isinstance(validator, (list, tuple, set)):
    return Options(*tuple(validator))
  if isinstance(validator, Validator):
    return validator
  else:
    raise AttributeDefinitionError('%s is not a valid validator' %
                                   str(validator))


def _SimplifiedValue(validator, value):
  """Convert any value to simplified collections and basic types.

  Args:
    validator: An instance of `Validator` that corresponds with 'value'.
      May also be 'str' or 'int' if those were used instead of a full
      Validator.
    value: Value to convert to simplified collections.

  Returns:
    The value as a dictionary if it is a `ValidatedBase` object.  A list of
    items converted to simplified collections if value is a list
    or a tuple. Otherwise, just the value.
  """
  if isinstance(value, ValidatedBase):

    return value.ToDict()
  elif isinstance(value, (list, tuple)):

    return [_SimplifiedValue(validator, item) for item in value]
  elif isinstance(validator, Validator):
    return validator.ToValue(value)
  return value


class ValidatedBase(object):
  """Base class for all validated objects."""

  @classmethod
  def GetValidator(cls, key):
    """Safely get the Validator corresponding to the given key.

    This function should be overridden by subclasses

    Args:
      key: The attribute or item to get a validator for.

    Returns:
      Validator associated with key or attribute.

    Raises:
      ValidationError: If the requested key is illegal.
    """
    raise NotImplementedError('Subclasses of ValidatedBase must '
                              'override GetValidator.')

  def SetMultiple(self, attributes):
    """Set multiple values on Validated instance.

    All attributes will be validated before being set.

    Args:
      attributes: A dict of attributes/items to set.

    Raises:
      ValidationError: When no validated attribute exists on class.
    """
    for key, value in attributes.items():
      self.Set(key, value)

  def Set(self, key, value):
    """Set a single value on Validated instance.

    This method should be overridded by sub-classes.

    This method can only be used to assign validated attributes/items.

    Args:
      key: The name of the attributes
      value: The value to set

    Raises:
      ValidationError: When no validated attribute exists on class.
    """
    raise NotImplementedError('Subclasses of ValidatedBase must override Set.')

  def CheckInitialized(self):
    """Checks for missing or conflicting attributes.

    Subclasses should override this function and raise an exception for
    any errors. Always run this method when all assignments are complete.

    Raises:
      ValidationError: When there are missing or conflicting attributes.
    """

  def ToDict(self):
    """Convert `ValidatedBase` object to a dictionary.

    Recursively traverses all of its elements and converts everything to
    simplified collections.

    Subclasses should override this method.

    Returns:
      A dictionary mapping all attributes to simple values or collections.
    """
    raise NotImplementedError('Subclasses of ValidatedBase must '
                              'override ToDict.')

  def ToYAML(self):
    """Print validated object as simplified YAML.

    Returns:
      Object as a simplified YAML string compatible with parsing using the
      `SafeLoader`.
    """
    return yaml.dump(self.ToDict(),
                     default_flow_style=False,
                     Dumper=ItemDumper)

  def GetWarnings(self):
    """Return all the warnings we've got, along with their associated fields.

    Returns:
      A list of tuples of (dotted_field, warning), both strings.
    """
    raise NotImplementedError('Subclasses of ValidatedBase must '
                              'override GetWarnings')


class Validated(ValidatedBase):
  """Base class for classes that require validation.

  A class which intends to use validated fields should sub-class itself from
  this class. Each class should define an 'ATTRIBUTES' class variable which
  should be a map from attribute name to its validator. For example:

    ```
    class Story(Validated):
      ATTRIBUTES = {'title': Type(str),
                    'authors': Repeated(Type(str)),
                    'isbn': Optional(Type(str)),
                    'pages': Type(int),
                    }
      ```

  Attributes that are not listed under `ATTRIBUTES` work like normal and are
  not validated upon assignment.
  """


  ATTRIBUTES = None

  def __init__(self, **attributes):
    """Constructor for Validated classes.

    This constructor can optionally assign values to the class via its
    keyword arguments.

    Raises:
      AttributeDefinitionError: When class instance is missing `ATTRIBUTE`
        definition or when `ATTRIBUTE` is of the wrong type.
    """
    super(Validated, self).__init__()
    if not isinstance(self.ATTRIBUTES, dict):
      raise AttributeDefinitionError(
          'The class %s does not define an ATTRIBUTE variable.'
          % self.__class__)


    for key in self.ATTRIBUTES.keys():
      object.__setattr__(self, key, self.GetValidator(key).default)

    self.SetMultiple(attributes)

  @classmethod
  def GetValidator(cls, key):
    """Safely get the underlying attribute definition as a `Validator`.

    Args:
      key: Name of attribute to get.

    Returns:
      Validator associated with key or attribute value wrapped in a
      validator.
    Raises:
      ValidationError: if no such attribute exists.
    """
    if key not in cls.ATTRIBUTES:
      raise ValidationError(
          'Unexpected attribute \'%s\' for object of type %s.' %
          (key, cls.__name__))

    return AsValidator(cls.ATTRIBUTES[key])

  def GetWarnings(self):
    ret = []
    for key in self.ATTRIBUTES.keys():
      ret.extend(self.GetValidator(key).GetWarnings(
          self.GetUnnormalized(key), key, self))
    return ret

  def Set(self, key, value):
    """Set a single value on `Validated` instance.

    This method can only be used to assign validated attributes.

    Args:
      key: The name of the attributes
      value: The value to set

    Raises:
      `ValidationError` when no validated attribute exists on class.
    """
    setattr(self, key, value)

  def GetUnnormalized(self, key):
    """Get a single value on the `Validated` instance, without normalizing."""
    validator = self.GetValidator(key)
    try:
      return super(Validated, self).__getattribute__(key)
    except AttributeError:
      return validator.default

  def Get(self, key):
    """Get a single value on Validated instance.

    This method can only be used to retrieve validated attributes.

    Args:
      key: The name of the attributes

    Raises:
      `ValidationError` when no validated attribute exists on class.
    """
    self.GetValidator(key)
    return getattr(self, key)

  def __getattribute__(self, key):



    ret = super(Validated, self).__getattribute__(key)
    if key in ['ATTRIBUTES', 'GetValidator', '__name__', '__class__']:
      return ret
    try:
      validator = self.GetValidator(key)
    except ValidationError:
      return ret
    if isinstance(validator, Normalized):
      return validator.Get(ret, key, self)
    return ret

  def CheckInitialized(self):
    for key in self.ATTRIBUTES.keys():
      value = self.GetUnnormalized(key)
      self.GetValidator(key).CheckFieldInitialized(value, key, self)

  def __setattr__(self, key, value):
    """Set attribute.

    Setting a value on an object of this type will only work for attributes
    defined in `ATTRIBUTES`. To make other assignments possible it is necessary
    to override this method in subclasses.

    It is important that assignment is restricted in this way because
    this validation is used as validation for parsing. Absent this restriction
    it would be possible for method names to be overwritten.

    Args:
      key: Name of attribute to set.
      value: The attribute's new value or None to unset.

    Raises:
      ValidationError: When trying to assign to an attribute
        that does not exist.
    """
    if value is not None:
      value = self.GetValidator(key)(value, key)
    object.__setattr__(self, key, value)

  def __str__(self):
    """Formatted view of validated object and nested values."""
    return repr(self)

  def __repr__(self):
    """Formatted view of validated object and nested values."""
    values = [(attr, getattr(self, attr)) for attr in self.ATTRIBUTES.keys()]
    dent = '    '
    value_list = []
    for attr, value in values:
      value_list.append('\n%s%s=%s' % (dent, attr, value))

    return "<%s %s\n%s>" % (self.__class__.__name__, ' '.join(value_list), dent)

  def __eq__(self, other):
    """Equality operator.

    Comparison is done by comparing all attribute values to those in the other
    instance. Objects which are not of the same type are not equal.

    Args:
      other: Other object to compare against.

    Returns:
      `True` if validated objects are equal, else `False`.
    """
    if type(self) != type(other):
      return False
    for key in self.ATTRIBUTES.keys():
      if getattr(self, key) != getattr(other, key):
        return False
    return True

  def __ne__(self, other):
    """Inequality operator."""
    return not self.__eq__(other)

  def __hash__(self):
    """Hash function for using Validated objects in sets and maps.

    Hash is done by hashing all keys and values and xor'ing them together.

    Returns:
      Hash of validated object.
    """
    result = 0
    for key in self.ATTRIBUTES.keys():
      value = getattr(self, key)
      if isinstance(value, list):
        value = tuple(value)
      result = result ^ hash(key) ^ hash(value)
    return result

  def ToDict(self):
    """Convert Validated object to a dictionary.

    Recursively traverses all of its elements and converts everything to
    simplified collections.

    Returns:
      A dict of all attributes defined in this classes `ATTRIBUTES` mapped
      to its value. This structure is recursive in that Validated objects
      that are referenced by this object and in lists are also converted to
      dicts.
    """
    result = {}
    for name, validator in self.ATTRIBUTES.items():
      value = self.GetUnnormalized(name)

      if not(isinstance(validator, Validator) and value == validator.default):
        result[name] = _SimplifiedValue(validator, value)
    return result


class ValidatedDict(ValidatedBase, dict):
  """Base class for validated dictionaries.

  You can control the keys and values that are allowed in the dictionary
  by setting `KEY_VALIDATOR` and `VALUE_VALIDATOR` to subclasses of `Validator`
  (or things that can be interpreted as validators, see `AsValidator`).

  For example if you wanted only capitalized keys that map to integers
  you could do:

    ```
    class CapitalizedIntegerDict(ValidatedDict):
      KEY_VALIDATOR = Regex('[A-Z].*')
      VALUE_VALIDATOR = int  # this gets interpreted to Type(int)
    ```

  The following code would result in an error:

    ```
    my_dict = CapitalizedIntegerDict()
    my_dict['lowercase'] = 5  # Throws a validation exception
    ```

  You can freely nest Validated and `ValidatedDict` inside each other so:

    ```
    class MasterObject(Validated):
      ATTRIBUTES = {'paramdict': CapitalizedIntegerDict}
    ```

  Could be used to parse the following YAML:

    ```
    paramdict:
      ArbitraryKey: 323
      AnotherArbitraryKey: 9931
    ```
  """
  KEY_VALIDATOR = None
  VALUE_VALIDATOR = None

  def __init__(self, **kwds):
    """Construct a validated dict by interpreting the key and value validators.

    Args:
      **kwds: keyword arguments will be validated and put into the dict.
    """
    super(ValidatedDict, self).__init__()
    self.update(kwds)

  @classmethod
  def GetValidator(cls, key):
    """Check the key for validity and return a corresponding value validator.

    Args:
      key: The key that will correspond to the validator we are returning.
    """
    key = AsValidator(cls.KEY_VALIDATOR)(key, 'key in %s' % cls.__name__)
    return AsValidator(cls.VALUE_VALIDATOR)

  def __setitem__(self, key, value):
    """Set an item.

    Only attributes accepted by `GetValidator` and values that validate
    with the validator returned from `GetValidator` are allowed to be set
    in this dictionary.

    Args:
      key: Name of item to set.
      value: Items new value.

    Raises:
      ValidationError: When trying to assign to a value that does not exist.
    """
    dict.__setitem__(self, key, self.GetValidator(key)(value, key))

  def setdefault(self, key, value=None):
    """Trap setdefaultss to ensure all key/value pairs are valid.

    See the documentation for setdefault on dict for usage details.

    Raises:
      ValidationError: if the specified key is illegal or the
      value invalid.
    """
    return dict.setdefault(self, key, self.GetValidator(key)(value, key))

  def update(self, other, **kwds):
    """Trap updates to ensure all key/value pairs are valid.

    See the documentation for update on dict for usage details.

    Raises:
      ValidationError: if any of the specified keys are illegal or
        values invalid.
    """
    if hasattr(other, 'keys') and callable(getattr(other, 'keys')):
      newother = {}
      for k in other:
        newother[k] = self.GetValidator(k)(other[k], k)
    else:
      newother = [(k, self.GetValidator(k)(v, k)) for (k, v) in other]

    newkwds = {}
    for k in kwds:
      newkwds[k] = self.GetValidator(k)(kwds[k], k)

    dict.update(self, newother, **newkwds)

  def Set(self, key, value):
    """Set a single value on `Validated` instance.

    This method checks that a given key and value are valid and if so
    puts the item into this dictionary.

    Args:
      key: The name of the attributes.
      value: The value to set.

    Raises:
      ValidationError: When no validated attribute exists on class.
    """
    self[key] = value

  def GetWarnings(self):
    ret = []
    for name, value in self.items():
      ret.extend(self.GetValidator(name).GetWarnings(value, name, self))
    return ret

  def ToDict(self):
    """Convert `ValidatedBase` object to a dictionary.

    Recursively traverses all of its elements and converts everything to
    simplified collections.

    Subclasses should override this method.

    Returns:
      A dictionary mapping all attributes to simple values or collections.
    """
    result = {}
    for name, value in self.items():
      validator = self.GetValidator(name)
      result[name] = _SimplifiedValue(validator, value)
    return result






class Validator(object):
  """Validator base class.

  Though any callable can be used as a validator, this class encapsulates the
  case when a specific validator needs to hold a particular state or
  configuration.

  To implement `Validator` sub-class, override the validate method.

  This class is permitted to change the ultimate value that is set to the
  attribute if there is a reasonable way to perform the conversion.
  """

  expected_type = object

  def __init__(self, default=None):
    """Constructor.

    Args:
      default: Default assignment is made during initialization and will
        not pass through validation.
    """
    self.default = default

  def __call__(self, value, key='???'):
    """Main interface to validator is call mechanism."""
    return self.Validate(value, key)

  def Validate(self, value, key='???'):
    """Validate this field. Override to customize subclass behavior.

    Args:
      value: Value to validate.
      key: Name of the field being validated.

    Returns:
      Value if value is valid, or a valid representation of value.
    """
    return value

  def CheckFieldInitialized(self, value, key, obj):
    """Check for missing fields or conflicts between fields.

    Default behavior performs a simple `None`-check, but this can be overridden.
    If the intent is to allow optional fields, then use the `Optional` validator
    instead.

    Args:
      value: Value to validate.
      key: Name of the field being validated.
      obj: The object to validate against.

    Raises:
      ValidationError: When there are missing or conflicting fields.
    """
    if value is None:
      raise MissingAttribute(key)

  def ToValue(self, value):
    """Convert `value` to a simplified collection or basic type.

    Subclasses of `Validator` should override this method when the dumped
    representation of `value` is not a simple `<type>(value)` (e.g., a regex).

    Args:
      value: An object of the same type that was returned from `Validate()`.

    Returns:
      An instance of a builtin type (e.g., int, str, dict, etc). By default
      it returns `value` unmodified.
    """
    return value

  def GetWarnings(self, value, key, obj):
    """Return any warnings on this attribute.

    Validates the value with an eye towards things that aren't fatal problems.

    Args:
      value: Value to validate.
      key: Name of the field being validated.
      obj: The object to validate against.

    Returns:
      A list of tuples (context, warning) where
        - Context is the field (or dotted field path, if a sub-field)
        - Warning is the string warning text
    """
    del value, key, obj
    return []


class StringValidator(Validator):
  """Verifies property is a valid text string.

  - In python 2: inherits from basestring.
  - In python 3: inherits from str.
  """

  def Validate(self, value, key='???'):
    if not isinstance(value, six.string_types):
      raise ValidationError(
          'Value %r for %s is not a valid text string.' % (
              value, key))
    return value


class Type(Validator):
  """Verifies property is of expected type.

  Can optionally convert value if it is not of the expected type.

  It is possible to specify a required field of a specific type in shorthand
  by merely providing the type. This method is slightly less efficient than
  providing an explicit type but is not significant unless parsing a large
  amount of information:

    ```
    class Person(Validated):
      ATTRIBUTES = {'name': unicode,
                    'age': int,
                    }
    ```

  However, in most instances it is best to use the type constants:

    ```
    class Person(Validated):
      ATTRIBUTES = {'name': TypeUnicode,
                    'age': TypeInt,
                    }
    ```
  """

  def __init__(self, expected_type, convert=True, default=None):
    """Initialize Type validator.

    Args:
      expected_type: Type that attribute should validate against.
      convert: Cause conversion if value is not the right type.
        Conversion is done by calling the constructor of the type
        with the value as its first parameter.
      default: Default assignment is made during initialization and will
        not pass through validation.
    """
    super(Type, self).__init__(default)
    self.expected_type = expected_type
    self.convert = convert

  def Validate(self, value, key):
    """Validate that `value` has the correct type.

    Args:
      value: Value to validate.
      key: Name of the field being validated.

    Returns:
      `value` if value is of the correct type. `value` is coverted to the
      correct type if the `Validator` is configured to do so.

    Raises:
      ValidationError: If `value` is not of the right type and the validator
        is either configured not to convert or cannot convert.
    """
    if isinstance(value, self.expected_type):
      return value





    if self.expected_type is str and isinstance(value, six.string_types):
      return value

    if not self.convert:
      raise ValidationError('Value %r for %s is not of the expected type %s' %
                            (value, key, self.expected_type.__name__))

    try:
      return self.expected_type(value)
    except ValueError as e:
      raise ValidationError(
          'Value %r for %s could not be converted to type %s.' %
          (value, key, self.expected_type.__name__), e)
    except TypeError as e:
      raise ValidationError(
          'Value %r for %s is not of the expected type %s' %
          (value, key, self.expected_type.__name__), e)

  def GetWarnings(self, value, key, obj):
    del obj
    if issubclass(self.expected_type, ValidatedBase):
      return [('%s.%s' % (key, subkey), warning)
              for subkey, warning in value.GetWarnings()]
    return []


TYPE_BOOL = Type(bool)
TYPE_INT = Type(int)




TYPE_LONG = Type(int)
TYPE_STR = Type(str)
TYPE_UNICODE = Type(six.text_type)
TYPE_FLOAT = Type(float)


class Exec(Type):
  """Coerces the value to accommodate Docker `CMD/ENTRYPOINT` requirements.

  Validates the value is a string, then tries to modify the string (if
  necessary) so that the command represented will become `PID` 1 inside the
  Docker container. See Docker documentation on `docker kill` for more
  information: https://docs.docker.com/engine/reference/commandline/kill/.

  If the command already starts with `exec` or appears to be in "exec form"
  (starts with `[`), no further action is needed. Otherwise, prepend the
  command with `exec` so that it will become `PID 1` on execution.
  """

  def __init__(self, default=None):
    """Initialize parent, a converting type validator for `str`."""
    super(Exec, self).__init__(str, convert=True, default=default)

  def Validate(self, value, key):
    """Validate according to parent behavior and coerce to start with `exec`."""


    value = super(Exec, self).Validate(value, key)

    if value.startswith('[') or value.startswith('exec'):

      return value
    else:

      return 'exec ' + value


class Options(Validator):
  """Limit field based on pre-determined values.

  Options are used to make sure an enumerated set of values are the only
  one permitted for assignment.  It is possible to define aliases which
  map multiple string values to a single original.  An example of usage:

    ```
    class ZooAnimal(validated.Class):
      ATTRIBUTES = {
        'name': str,
        'kind': Options('platypus',                   # No aliases
                        ('rhinoceros', ['rhino']),    # One alias
                        ('canine', ('dog', 'puppy')), # Two aliases
                        )
    ```
  """

  def __init__(self, *options, **kw):
    """Initialize options.

    Args:
      options: List of allowed values.
    """
    if 'default' in kw:
      default = kw['default']
    else:
      default = None

    alias_map = {}
    def AddAlias(alias, original):
      """Set new alias on `alias_map`.

      Raises:
        AttributeDefinitionError: When option already exists or if alias is
          not of type str.
      """
      if not isinstance(alias, str):
        raise AttributeDefinitionError(
            'All option values must be of type str.')
      elif alias in alias_map:
        raise AttributeDefinitionError(
            "Option '%s' already defined for options property." % alias)
      alias_map[alias] = original


    for option in options:
      if isinstance(option, str):
        AddAlias(option, option)

      elif isinstance(option, (list, tuple)):

        if len(option) != 2:
          raise AttributeDefinitionError("Alias is defined as a list or tuple "
                                         "with two items.  The first is the "
                                         "original option, while the second "
                                         "is a list or tuple of str aliases.\n"
                                         "\n  Example:\n"
                                         "      ('original', ('alias1', "
                                         "'alias2'")
        original, aliases = option
        AddAlias(original, original)
        if not isinstance(aliases, (list, tuple)):
          raise AttributeDefinitionError('Alias lists must be a list or tuple')

        for alias in aliases:
          AddAlias(alias, original)

      else:
        raise AttributeDefinitionError("All options must be of type str "
                                       "or of the form (str, [str...]).")
    super(Options, self).__init__(default)
    self.options = alias_map

  def Validate(self, value, key):
    """Validate options.

    Returns:
      Original value for provided alias.

    Raises:
      ValidationError: When the value is not one of predefined values.
    """
    value = str(value)
    if value not in self.options:
      raise ValidationError('Value \'%s\' for %s not in %s.'
                            % (value, key, self.options))
    return self.options[value]


class Optional(Validator):
  """Definition of optional attributes.

  Optional values are attributes which can be set to `None` or left
  unset. All values in a basic `Validated` class are set to `None`
  at initialization. Failure to assign to non-optional values
  will result in a validation error when calling `CheckInitialized`.
  """

  def __init__(self, validator, default=None):
    """Initializer.

    This constructor will make a few guesses about the value passed in
    as the validator:

      - If the validator argument is a type, it automatically creates a Type
        validator around it.

      - If the validator argument is a list or tuple, it automatically
        creates an Options validator around it.

    Args:
      validator: Optional validation condition.

    Raises:
      AttributeDefinitionError: if validator is not callable.
    """
    self.validator = AsValidator(validator)
    self.expected_type = self.validator.expected_type
    self.default = default

  def Validate(self, value, key):
    """Optionally require a value.

    Normal validators do not accept `None`. This will accept `None` on
    behalf of the contained validator.

    Args:
      value: Value to be validated as optional.
      key: Name of the field being validated.

    Returns:
      None if value is `None`, else results of contained validation.
    """
    return self.validator(value, key)

  def CheckFieldInitialized(self, value, key, obj):
    if value is None:
      return
    self.validator.CheckFieldInitialized(value, key, obj)

  def ToValue(self, value):
    """Convert `value` to a simplified collection or basic type."""
    if value is None:
      return None
    return self.validator.ToValue(value)


class Regex(Validator):
  """Regular expression validator.

  Regular expression validator always converts value to string. Note that
  matches must be exact.  Partial matches will not validate. For example:

    class ClassDescr(Validated):
      ATTRIBUTES = { 'name': Regex(r'[a-zA-Z_][a-zA-Z_0-9]*'),
                     'parent': Type(type)`
                     }

  Alternatively, any attribute that is defined as a string is automatically
  interpreted to be of type Regex.  It is possible to specify unicode regex
  strings as well.  This approach is slightly less efficient, but usually
  is not significant unless parsing large amounts of data:

    ```
    class ClassDescr(Validated):
      ATTRIBUTES = { 'name': r'[a-zA-Z_][a-zA-Z_0-9]*',
                     'parent': Type(type),
                     }

    # This will raise a ValidationError exception.
    my_class(name='AName with space', parent=AnotherClass)
    ```
  """

  def __init__(self, regex, string_type=six.text_type, default=None):
    """Initialized regex validator.

    Args:
      regex: Regular expression string to use for comparison.
      string_type: Type to be considered a string.
      default: Default value.

    Raises:
      AttributeDefinitionError: if `string_type` is not a kind of string.
    """
    super(Regex, self).__init__(default)
    if (not issubclass(string_type, six.string_types) or
        _IsPy2Basestring(string_type)):
      raise AttributeDefinitionError(
          'Regex fields must be a string type not %s.' % str(string_type))
    if isinstance(regex, six.string_types):
      self.re = re.compile('^(?:%s)$' % regex)
    else:
      raise AttributeDefinitionError(
          'Regular expression must be string.  Found %s.' % str(regex))

    self.expected_type = string_type

  def Validate(self, value, key):
    """Does validation of a string against a regular expression.

    Args:
      value: String to match against regular expression.
      key: Name of the field being validated.

    Raises:
      ValidationError: When value does not match regular expression or
        when value does not match provided string type.
    """
    if issubclass(self.expected_type, str):
      cast_value = TYPE_STR(value)
    else:
      cast_value = TYPE_UNICODE(value)

    if self.re.match(cast_value) is None:
      raise ValidationError('Value \'%s\' for %s does not match expression '
                            '\'%s\'' % (value, key, self.re.pattern))
    return cast_value


class _RegexStrValue(object):
  """Simulates the regex object to support recompilation when necessary.

  Used by the `RegexStr` class to dynamically build and recompile regular
  expression attributes of a validated object. This object replaces the normal
  object returned from re.compile which is immutable.

  When the value of this object is a string, that string is simply used as the
  regular expression when recompilation is needed. If the state of this object
  is a list of strings, the strings are joined in to a single `or` expression.
  """

  def __init__(self, attribute, value, key):
    """Initialize recompilable regex value.

    Args:
      attribute: Attribute validator associated with this regex value.
      value: Initial underlying python value for regex string. Either a single
        regex string or a list of regex strings.
      key: Name of the field.
    """
    self.__attribute = attribute
    self.__value = value
    self.__regex = None
    self.__key = key

  def __AsString(self, value):
    """Convert a value to appropriate string.

    Returns:
      String version of value with all carriage returns and line feeds removed.
    """
    if issubclass(self.__attribute.expected_type, str):
      cast_value = TYPE_STR(value)
    else:
      cast_value = TYPE_UNICODE(value)

    cast_value = cast_value.replace('\n', '')
    cast_value = cast_value.replace('\r', '')
    return cast_value

  def __BuildRegex(self):
    """Build regex string from state.

    Returns:
      String version of regular expression. Sequence objects are constructed
      as larger regular expression where each regex in the list is joined with
      all the others as single `or` expression.
    """
    if isinstance(self.__value, list):
      value_list = self.__value
      sequence = True
    else:
      value_list = [self.__value]
      sequence = False

    regex_list = []
    for item in value_list:
      regex_list.append(self.__AsString(item))

    if sequence:
      return '|'.join('%s' % item for item in regex_list)
    else:
      return regex_list[0]

  def __Compile(self):
    """Build regular expression object from state.

    Returns:
      Compiled regular expression based on internal value.
    """
    regex = self.__BuildRegex()
    try:
      return re.compile(regex)
    except re.error as e:
      raise ValidationError('Value \'%s\' for %s does not compile: %s' %
                            (regex, self.__key, e), e)

  @property
  def regex(self):
    """Compiled regular expression as described by underlying value."""
    return self.__Compile()

  def match(self, value):
    """Match against internal regular expression.

    Args:
      value: String to match against regular expression.

    Returns:
      Regular expression object built from underlying value.
    """
    return re.match(self.__BuildRegex(), value)

  def Validate(self):
    """Ensure that regex string compiles."""
    self.__Compile()

  def __str__(self):
    """Regular expression string as described by underlying value."""
    return self.__BuildRegex()

  def __eq__(self, other):
    """Comparison against other regular expression string values."""
    if isinstance(other, _RegexStrValue):
      return self.__BuildRegex() == other.__BuildRegex()
    return str(self) == other

  def __ne__(self, other):
    """Inequality operator for regular expression string value."""
    return not self.__eq__(other)


class RegexStr(Validator):
  """Validates that a string can compile as a regex without errors.

  Use this validator when the value of a field should be a regex. That
  means that the value must be a string that can be compiled by `re.compile()`.
  The attribute will then be a compiled re object.
  """

  def __init__(self, string_type=six.text_type, default=None):
    """Initialized regex validator.

    Args:
      string_type: Type to be considered a string.
      default: Default value.

    Raises:
      AttributeDefinitionError: If `string_type` is not a kind of string.
    """
    if default is not None:
      default = _RegexStrValue(self, default, None)
      re.compile(str(default))
    super(RegexStr, self).__init__(default)



    if (not issubclass(string_type, six.string_types) or
        _IsPy2Basestring(string_type)):
      raise AttributeDefinitionError(
          'RegexStr fields must be a string type not %s.' % str(string_type))

    self.expected_type = string_type

  def Validate(self, value, key):
    """Validates that the string compiles as a regular expression.

    Because the regular expression might have been expressed as a multiline
    string, this function also strips newlines out of value.

    Args:
      value: String to compile as a regular expression.
      key: Name of the field being validated.

    Raises:
      `ValueError` when value does not compile as a regular expression.
      `TypeError` when value does not match provided string type.
    """
    if isinstance(value, _RegexStrValue):
      return value
    value = _RegexStrValue(self, value, key)
    value.Validate()
    return value

  def ToValue(self, value):
    """Returns the `RE` pattern for this validator."""
    return str(value)


class Range(Validator):
  """Validates that numbers fall within the correct range.

  In theory this class can be emulated using Options, however error
  messages generated from that class will not be very intelligible.
  This class essentially does the same thing, but knows the intended
  integer range.

  Also, this range class supports floats and other types that implement
  ordinality.

  The range is inclusive, meaning 3 is considered in the range
  in `Range(1,3)`.
  """

  def __init__(self, minimum, maximum, range_type=int, default=None):
    """Initializer for range.

    At least one of minimum and maximum must be supplied.

    Args:
      minimum: Minimum for attribute.
      maximum: Maximum for attribute.
      range_type: Type of field. Defaults to int.

    Raises:
      AttributeDefinitionError: If the specified parameters are incorrect.
    """
    super(Range, self).__init__(default)
    min_max_type = range_type
    if range_type in six.integer_types:
      min_max_type = six.integer_types
    if minimum is None and maximum is None:
      raise AttributeDefinitionError('Must specify minimum or maximum.')
    if minimum is not None and not isinstance(minimum, min_max_type):
      raise AttributeDefinitionError(
          'Minimum value must be of type %s, instead it is %s (%s).' %
          (str(range_type), str(type(minimum)), str(minimum)))
    if maximum is not None and not isinstance(maximum, min_max_type):
      raise AttributeDefinitionError(
          'Maximum value must be of type %s, instead it is %s (%s).' %
          (str(range_type), str(type(maximum)), str(maximum)))

    self.minimum = minimum
    self.maximum = maximum
    self.expected_type = range_type
    self._type_validator = Type(range_type)

  def Validate(self, value, key):
    """Validate that value is within range.

    Validates against range-type then checks the range.

    Args:
      value: Value to validate.
      key: Name of the field being validated.

    Raises:
      ValidationError: When value is out of range.
    """
    cast_value = self._type_validator.Validate(value, key)
    if self.maximum is None and cast_value < self.minimum:
      raise ValidationError('Value \'%s\' for %s less than %s'
                            % (value, key, self.minimum))
    elif self.minimum is None and cast_value > self.maximum:
      raise ValidationError('Value \'%s\' for %s greater than %s'
                            % (value, key, self.maximum))

    elif ((self.minimum is not None and cast_value < self.minimum) or
          (self.maximum is not None and cast_value > self.maximum)):
      raise ValidationError('Value \'%s\' for %s is out of range %s - %s'
                            % (value, key, self.minimum, self.maximum))
    return cast_value


class Repeated(Validator):
  """Repeated field validator.

  Indicates that attribute is expected to be a repeated value, ie,
  a sequence. This adds additional validation over just `Type(list)`
  in that it retains information about what can be stored in the list by
  use of its constructor field.
  """

  def __init__(self, constructor, default=None):
    """Initializer for repeated field.

    Args:
      constructor: Type used for verifying elements of sequence attribute.
    """
    super(Repeated, self).__init__(default)
    self.constructor = constructor
    self.expected_type = list

  def Validate(self, value, key):
    """Do validation of sequence.

    Value must be a list and all elements must be of type `constructor`.

    Args:
      value: Value to validate.
      key: Name of the field being validated.

    Raises:
      ValidationError: if value is None, not a list or one of its elements is
        the wrong type.
    """
    if not isinstance(value, list):
      raise ValidationError('Value \'%s\' for %s should be a sequence but '
                            'is not.' % (value, key))

    for idx, item in enumerate(value):
      if isinstance(self.constructor, Validator):
        value[idx] = self.constructor.Validate(item, key)
      elif not isinstance(item, self.constructor):
        raise ValidationError('Value element \'%s\' for %s must be type %s.' % (
            str(item), key, self.constructor.__name__))

    return value

  def CheckFieldInitialized(self, value, key, obj):
    if value is None:
      raise MissingAttribute(key)
    for idx, item in enumerate(value):
      if isinstance(self.constructor, Validator):
        self.constructor.CheckFieldInitialized(item, key, self)


        value[idx] = self.constructor.Validate(item, key)
      elif not isinstance(item, self.constructor):
        raise ValidationError('Value element \'%s\' for %s must be type %s.' % (
            str(item), key, self.constructor.__name__))


class TimeValue(Validator):
  """Validates time values with units, such as 1h or 3.5d."""

  _EXPECTED_SYNTAX = ('must be a non-negative number followed by a time unit, '
                      'such as 1h or 3.5d')

  def __init__(self):
    super(TimeValue, self).__init__()
    self.expected_type = str

  def Validate(self, value, key):
    """Validate a time value.

    Args:
      value: Value to validate.
      key: Name of the field being validated.

    Raises:
      ValidationError: if value is not a time value with the expected format.
    """
    if not isinstance(value, six.string_types):
      raise ValidationError("Value '%s' for %s is not a string (%s)"
                            % (value, key, TimeValue._EXPECTED_SYNTAX))
    if not value:
      raise ValidationError("Value for %s is empty (%s)"
                            % (key, TimeValue._EXPECTED_SYNTAX))
    if value[-1] not in "smhd":
      raise ValidationError("Value '%s' for %s must end with a time unit, "
                            "one of s (seconds), m (minutes), h (hours), "
                            "or d (days)" % (value, key))
    try:
      t = float(value[:-1])
    except ValueError:
      raise ValidationError("Value '%s' for %s is not a valid time value (%s)"
                            % (value, key, TimeValue._EXPECTED_SYNTAX))
    if t < 0:
      raise ValidationError("Value '%s' for %s is negative (%s)"
                            % (value, key, TimeValue._EXPECTED_SYNTAX))
    return value


class Normalized(Validator):
  """Normalizes a field on lookup, but serializes with the original value.

  Only works with fields on `Validated`.
  """

  def Validate(self, value, key):
    return self.validator(value, key)

  def Get(self, value, key, obj):
    """Returns the normalized value. Subclasses must override."""
    raise NotImplementedError('Subclasses must override `Get`!')


class Preferred(Normalized):
  """A non-deprecated field when there's a deprecated one.

  For use with Deprecated. Only works as a field on `Validated`.

  Both fields will work for value access. It's an error to set both the
  deprecated and the corresponding preferred field.
  """

  def __init__(self, deprecated, validator, default=None):
    """Initializer for Preferred.

    Args:
      deprecated: The name of the corresponding deprecated field
      validator: The validator for the actual value of this field.
      default: The default value for this field.
    """
    super(Preferred, self).__init__(default=None)
    self.validator = AsValidator(validator)
    self.deprecated = deprecated


    self.synthetic_default = default

  def CheckFieldInitialized(self, value, key, obj):
    deprecated_value = obj.GetUnnormalized(self.deprecated)
    if value is not None and deprecated_value is not None:
      raise ValidationError('Only one of the two fields [{}] (preferred)'
                            ' and [{}] (deprecated) may be set.'.format(
                                key, self.deprecated))
    if deprecated_value is not None:
      return
    if not self.synthetic_default:
      self.validator.CheckFieldInitialized(value, key, obj)

  def Get(self, value, key, obj):
    if value is not None:
      return value
    deprecated_value = obj.GetUnnormalized(self.deprecated)
    if deprecated_value is not None:
      return deprecated_value
    return self.synthetic_default


class Deprecated(Normalized):
  """A deprecated field.

  For use with `Preferred`. Only works as a field on `Validated`.

  Both fields will work for value access. It's an error to set both the
  deprecated and the corresponding preferred field.
  """

  def __init__(self, preferred, validator, default=None):
    """Initializer for Deprecated.

    Args:
      preferred: The name of the preferred field.
      validator: The validator for the actual value of this field.
      default: The default value for this field.
    """
    super(Deprecated, self).__init__(default=None)
    self.validator = Optional(validator)
    self.preferred = preferred
    self.synthetic_default = default

  def GetWarnings(self, value, key, obj):
    del obj
    if value is not None:
      return [(
          key,
          'Field %s is deprecated; use %s instead.' % (key, self.preferred))]
    return []

  def Get(self, value, key, obj):
    preferred_value = obj.GetUnnormalized(self.preferred)
    if preferred_value is not None:
      return preferred_value
    elif value is not None:
      return value
    else:
      return self.synthetic_default

  def CheckFieldInitialized(self, value, key, obj):
    pass





def _IsPy2Basestring(x):
  return six.PY2 and x is basestring



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helpers providing a bridge between the Python 2 `__cmp__` and Python 3.

Python 2 uses methods called `__cmp__` to override object comparison behavior;
in Python 3 `__cmp__` is replaced by rich comparison functions.

Additionally, Python 2 allows code to compare objects of different types,
falling back to lexographical class-name comparisons if types are different (so
instance of "int" are always less than instances of "str").

This file offers some utilities to bridge the gap between Python 2 and 3.
"""


import functools
import six
from six.moves import zip


def total_ordering_from_cmp(cls):
  """Class decorator that fills in missing ordering methods from `__cmp__`.

  This lets us take a class defined for Python 2's ordering system (using
  `__cmp__`) and use it with minimal changes — just a decorator — in Python 3.

  This implementation is adapted from the Python 2 version of
  `functools.total_ordering`, which derives these methods from each other
  instead of from `__cmp__`.

  Args:
    cls: The class to decorate.

  Returns:
    The decorated class.
  """

  convert = [('__gt__', lambda self, other: self.__cmp__(other) > 0),
             ('__lt__', lambda self, other: self.__cmp__(other) < 0),
             ('__le__', lambda self, other: self.__cmp__(other) <= 0),
             ('__eq__', lambda self, other: self.__cmp__(other) == 0),
             ('__ne__', lambda self, other: self.__cmp__(other) != 0),
             ('__ge__', lambda self, other: self.__cmp__(other) >= 0)]

  for opname, opfunc in convert:
    opfunc.__name__ = opname
    opfunc.__doc__ = getattr(int, opname).__doc__
    setattr(cls, opname, opfunc)
  return cls


if six.PY2:
  cmp = cmp
else:

  def _cmp_fallback(a, b):
    """cmp fallback for types not directly comparable via __gt__ and __lt__."""

    if type(a) is type(b):

      return cmp(id(a), id(b))


    return cmp(str(type(a)), str(type(b)))

  def _cmp_gt_lt(a, b):
    """cmp using __gt__ and __lt__."""

    a_type = type(a)

    gt = a_type.__gt__(a, b)
    if gt is NotImplemented:
      return gt

    lt = a_type.__lt__(a, b)
    if lt is NotImplemented:
      return lt

    return gt - lt

  def cmp(a, b):
    """Emulate Python 2 cmp in Python 3.

    This is not necessarily a complete emulation; it's simply enough to support
    Titanoboa's port of apphosting.

    Args:
      a: An object to compare
      b: Another object to compare

    Returns:
      Negative if a is less than b, 0 if a equals b, and positive otherwise.
    """

    a_type = type(a)
    b_type = type(b)

    if a_type is b_type and isinstance(a, (list, tuple)):

      for a_elem, b_elem in zip(a, b):
        rc = cmp(a_elem, b_elem)
        if rc != 0:
          return rc
      return cmp(len(a), len(b))

    if a_type is b_type and isinstance(a, dict):

      a_sorted = sorted(six.iteritems(a), key=functools.cmp_to_key(cmp))
      b_sorted = sorted(six.iteritems(b), key=functools.cmp_to_key(cmp))
      return cmp(a_sorted, b_sorted)

    result = _cmp_gt_lt(a, b)
    if result is not NotImplemented:
      return result

    result = _cmp_gt_lt(b, a)
    if result is not NotImplemented:
      return -result

    return _cmp_fallback(a, b)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Errors used in the YAML API."""




class Error(Exception):
  """Base datastore yaml error type."""

class ProtocolBufferParseError(Error):
  """Error in protocol buffer parsing."""


class EmptyConfigurationFile(Error):
  """Tried to load an empty configuration file."""


class MultipleConfigurationFile(Error):
  """Tried to load configuration file with multiple objects."""


class AmbiguousConfigurationFiles(Error):
  """Both YAML and XML files exist for the same configuration information."""


class UnexpectedAttribute(Error):
  """Raised when an unexpected attribute is encounted."""


class DuplicateAttribute(Error):
  """Generated when an attribute is assigned twice."""


class ListenerConfigurationError(Error):
  """Generated when there is a parsing problem due to configuration."""


class IllegalEvent(Error):
  """Raised when an unexpected event type is received by the listener."""


class InternalError(Error):
  """Raised when an internal implementation error is detected."""


class EventListenerError(Error):
  """Top level exception raised by the YAML listener.

  Any exception raised within the process of parsing a YAML file from an
  `EventListener` is caught and wrapped in an `EventListenerError`. The causing
  exception is maintained, but additional useful information is saved which
  can be used for reporting useful information to users.

  Attributes:
    cause: The original exception which caused the `EventListenerError`.
  """

  def __init__(self, cause):
    """Initialize event-listener error."""
    if hasattr(cause, 'args') and cause.args:
      Error.__init__(self, *cause.args)
    else:


      Error.__init__(self, str(cause))
    self.cause = cause

  def __str__(self):
    return str(self.cause)


class EventListenerYAMLError(EventListenerError):
  """Generated specifically for `yaml.error.YAMLError`."""


class EventError(EventListenerError):
  """Generated specifically when an error occurs in the event handler.

  Attributes:
    cause: The original exception which caused the `EventListenerError`.
    event: Event being handled when exception occured.
  """

  def __init__(self, cause, event):
    """Initialize event-listener error."""
    EventListenerError.__init__(self, cause)
    self.event = event

  def __str__(self):
    return '%s\n%s' % (self.cause, self.event.start_mark)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Functions that prepare GAE user code for running in a GCE VM."""


import functools
import json
import logging
import math
import os
import re
import typing
import wsgiref.util





from google.appengine.runtime import default_api_stub
from google.appengine.runtime import thread_hooks

import six


def GetTraceAndSpanId():
  """Gets the Trace and Span ID for the given request via environment variables.

  Checks and parses the envirionment variable "HTTP_X_CLOUD_TRACE_CONTEXT",
  which is a header provided by AppEngine (and is set as an environment
  variable as a part of the WSGI Specification).

  Returns:
    A tuple of Trace ID and the Span ID
  """


  ctx = os.getenv('HTTP_X_CLOUD_TRACE_CONTEXT')
  if ctx:
    m = re.search(r'^(\w+)/(\d+)(?:;o=[01])?$', ctx)
    if not m:
      return (None, None)
    trace_id = m.group(1)
    span_id = m.group(2)
    project_id = os.getenv('GOOGLE_CLOUD_PROJECT')
    return ('projects/{}/traces/{}'.format(project_id, trace_id), span_id)
  return (None, None)


def GetRequestUrl():
  """Constructs the full request url from WSGI environment variables."""
  try:
    environ = typing.cast(typing.Dict[str, typing.Any], os.environ)
    return wsgiref.util.request_uri(environ)
  except KeyError:
    return None


class JsonFormatter(logging.Formatter):
  """Class for logging to the cloud logging api with json metadata."""

  def format(self, record):
    """Format the record as json the cloud logging agent understands.

    Args:
      record: A logging.LogRecord to format.

    Returns:
      A json string to log.
    """
    float_frac_sec, float_sec = math.modf(record.created)

    message = record.getMessage()
    if record.exc_info:
      message = '%s\n%s' % (message, self.formatException(record.exc_info))

    data = {
        'message': message,
        'thread': record.thread,
        'severity': record.levelname,
        'timestamp': {
            'seconds': int(float_sec),
            'nanos': int(float_frac_sec * 1000000000)
        },
        'logging.googleapis.com/sourceLocation': {
            'file': record.pathname,
            'line': str(record.lineno),
            'function': record.funcName,
        },
        'serviceContext': {
            'version': os.getenv('GAE_VERSION'),
            'service': os.getenv('GAE_SERVICE'),
        },
        'context': {
            'user': os.getenv('USER_NICKNAME'),
            'httpRequest': {
                'url': GetRequestUrl(),
                'userAgent': os.getenv('HTTP_USER_AGENT'),
                'requestMethod': os.getenv('REQUEST_METHOD'),
                'protocol': os.getenv('SERVER_PROTOCOL')
            }
        },
    }

    trace_id, span_id = GetTraceAndSpanId()
    if trace_id:
      data['logging.googleapis.com/trace'] = trace_id
      data['logging.googleapis.com/spanId'] = span_id

    return json.dumps(data)


class SplitFileHandler(logging.FileHandler):
  """Class for splitting large logs into chunks."""

  def emit(self, record):
    """Emit a record.

    If the message is larger than the max size of a log entry of 256KB
    (https://cloud.google.com/logging/quotas#log-limits), it's split in
    chunks to prevent it from being lost

    Args:
      record: an instance of logging.LogRecord
    """
    message = str(record.msg)


    max_message_size = 256000
    if len(message) <= max_message_size or six.PY2:
      super(SplitFileHandler, self).emit(record)
    else:
      chunks = [
          message[i:i + max_message_size]
          for i in range(0, len(message), max_message_size)
      ]

      for idx, chunk in enumerate(chunks):
        record.msg = 'Part {}/{}: {}'.format(str(idx + 1), len(chunks), chunk)
        super().emit(record)


def InitializeFileLogging(log_path, clear_logging_handlers,
                          custom_json_formatter=None):
  """Helper called from CreateAndRunService() to set up syslog logging."""




  logging.basicConfig()

  logger = logging.getLogger()

  if clear_logging_handlers:



    if len(logger.handlers) > 1:
      logger.warning(
          'Removing more than one logging handler. '
          'This implies that a user-added logging handler is being removed!')
    logger.handlers[:] = []





  file_handler = SplitFileHandler(log_path)
  json_formatter = custom_json_formatter or JsonFormatter()
  file_handler.setFormatter(json_formatter)
  logger.addHandler(file_handler)

  logger.setLevel(logging.DEBUG)


class SecurityTicketThreadHook(thread_hooks.ThreadHook):
  """Sets and clears UseRequestSecurityTicket on a thread."""

  def PreTarget(self):
    default_api_stub.DefaultApiStub.SetUseRequestSecurityTicketForThread(True)

  def PostTarget(self):
    default_api_stub.DefaultApiStub.SetUseRequestSecurityTicketForThread(False)


@functools.lru_cache(maxsize=None)
def InitializeThreadingApis():
  """Helper to monkey-patch various threading APIs."""


  thread_hooks.PatchStartNewThread(hooks=[
      SecurityTicketThreadHook, thread_hooks.RequestEnvironmentThreadHook
  ])
